//
// Copyright (C) 2010  Clement Creusot
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef COMMONH
#define COMMONH

#include <stdlib.h>
#include <string.h>
#include <sstream>
#include <iostream>
#include <limits>
#include <vector>
#include <string>
#include <map>
#include <set>
#include <list>
#include <fstream>



/*
 * Macro definitions.
 */ 

/*
 * Print basic error
 * with file and line where it occured
 */
#define PRINT_ERROR(reason, ...) do {					\
    char _str[100];							\
    sprintf(_str, reason, ##__VA_ARGS__);				\
    const char * at = strstr(__FILE__,"face/");				\
    if (at != NULL) {fprintf(stderr,"ERROR %s %d: %s\n",at,__LINE__,_str);} \
    else      {fprintf(stderr,"ERROR %s %d: %s\n",__FILE__,__LINE__,_str);} \
    exit(EXIT_FAILURE);                                                 \
  } while(0)   

/*
 * Print basic warning
 */
#define PRINT_WARNING(reason, ...) do {					\
    char _str[100];							\
    sprintf(_str, reason, ##__VA_ARGS__);				\
    const char * at = strstr(__FILE__,"face/");				\
    if (at != NULL) {fprintf(stderr,"WARNING %s %d: %s\n",at,__LINE__,_str);} \
    else      {fprintf(stderr,"WARNING %s %d: %s\n",__FILE__,__LINE__,_str);} \
  } while(0)   



/*
 * Check most of system call correct execution.
 */
#define STDCALL_CHECK(expr,reason) do {					\
  if ((int)(expr) == -1){                                               \
    perror(reason);                                                     \
    exit(EXIT_FAILURE);                                                 \
   }                                                                    \
  } while(0)   



//
// Set built-in type.  Creates member Set"name"() (e.g., SetVisibility());
//
#define SET_MACRO(type, name)	   \
virtual void Set##name (type _arg) \
  { \
  if (this->name != _arg) \
    { \
    this->name = _arg; \
    } \
  } 

//
// Get built-in type.  Creates member Get"name"() (e.g., GetVisibility());
//
#define GET_MACRO(type, name)    \
virtual type Get##name () { \
  return this->name; \
  } 



//
// Set character string.  Creates member Set"name"() 
// (e.g., SetFilename(char *));
//
#define SET_STRING_MACRO(name) \
void Set##name (const char* _arg) \
  { \
  if ( this->name == NULL && _arg == NULL) { return;} \
  if ( this->name && _arg && (!strcmp(this->name,_arg))) { return;} \
  if (this->name) { delete [] this->name; } \
  if (_arg) \
    { \
    size_t n = strlen(_arg) + 1; \
    char *cp1 =  new char[n]; \
    const char *cp2 = (_arg); \
    this->name = cp1; \
    do { *cp1++ = *cp2++; } while ( --n ); \
    } \
   else \
    { \
    this->name = NULL; \
    } \
  } 

//
// Get character string.  Creates member Get"name"() 
// (e.g., char *GetFilename());
//
#define GET_STRING_MACRO(name)			\
  char* Get##name () {				\
  return this->name; \
  } 


inline void PrintOctaveMatrix(std::string& outputName, const std::vector<std::vector<double> >& mat, int width=6, int precision=4){
  int nQuery = mat.size();
  if (nQuery == 0){
    return;
  }
  int nModel = mat[0].size();

  std::ofstream out;
  out.open(outputName.c_str(), std::fstream::out);    
  out << "# name: Y" << std::endl;
  out << "# type: matrix" << std::endl;
  out << "# rows: " << nQuery << std::endl;
  out << "# columns: " << nModel << std::endl;

  out.precision(precision);
  out.setf(std::ios::fixed,std::ios::floatfield); 

  for(int i=0; i<nQuery; i++){  
    for(int j=0; j<nModel; j++){
      out.width(width);
      out << mat[i][j] << " ";
    }
    out << std::endl;
  }
  out.close();
}



template <class M,class N>
  bool pairFirstCompare (std::pair<M,N> i, std::pair<M,N> j) { return (i.first<j.first); }

template <class M,class N>
  bool pairSecondCompare (std::pair<M,N> i, std::pair<M,N> j) { return (i.second<j.second); }


// print vector of double
inline std::ostream& operator<<(std::ostream& stream,std::vector<double>& v){
  for(unsigned int i=0; i<v.size(); i++){    
      if (i!=0){
	stream << " " ;
      }
      stream << v[i] ;
  }
  return stream;
}

// print vector of 3D position //cc dangerous but used only for debug
inline std::ostream& operator<<(std::ostream& stream,std::vector<double*>& vector){
  for(unsigned int i=0; i<vector.size(); i++){
    if (i!=0){
      stream << " " ;
    } 
    stream << vector[i][0]<<","<< vector[i][1]<<","<< vector[i][2];
  }
  return stream;
}



inline std::string FindCorrespondingFile(const std::string& name,
					 const char* suffix){
  std::string resname = name;
  size_t endpt = resname.size();
  while(1){
    endpt = name.substr(0,endpt).rfind(".");
    if (endpt == std::string::npos){
      break;
    }
    resname = name.substr(0,endpt)+suffix;
    std::ifstream in;
    in.open(resname.c_str());
    if(in) {
      in.close();
      break;
    }
    in.close();  
  }
  return resname;
}
  

// find a T in a map or print an error
template <class T>
inline T map_extract(std::map<std::string, T> map, std::string key)
{
  typename std::map<std::string,T>::const_iterator 
    it = map.find(key);
  if (it == map.end()){
    PRINT_ERROR("The field \"%s\" is not defined in this map.",key.c_str());
  }    
  return (*it).second; 
}

// find a T in a map or print an error
template <class T,class P>
  inline bool map_has_key(const std::map<T, P> map, const T &key){
  typename std::map<T,P>::const_iterator 
    it = map.find(key);
  if (it == map.end()){
    return false;
  }    
  return true;
}



template <class T>
inline T MapGet(const std::map<std::string, std::string>& map, 
		const std::string& str)
{
  const std::map<std::string, std::string>::const_iterator 
    it = map.find(str);
  if (it == map.end())
    PRINT_ERROR("No field named %s", str.c_str());
  std::stringstream ss((*it).second);
  T value;   
  ss >> value;
   return value; 
}



// convert to string almost anything
template <class stringable>
inline std::string to_string (const stringable& t)
{
  std::stringstream ss;
  ss << t;
  return ss.str();
}

// convert to string almost anything
template <class stringable>
inline std::string zfill (const stringable& t, int size)
{
  std::string word = to_string(t);
  int len = word.size();
  if ( len < size){
    word.insert(0,size-len,'0');
  }
  return word;
}


// convert to string almost anything
inline std::string append(const std::string& a, const std::string& b)
{
  std::stringstream ss;
  ss << a << b;
  return ss.str();
}




// for conveniance
template <class Precision>
inline void GetMeanAndDev(std::vector<Precision> tab, double& mean, double& deviation, 
			  Precision& minimum, Precision& maximum,
			  unsigned int* argmin=NULL, unsigned int* argmax=NULL,
			  Precision lowerbound=-std::numeric_limits<Precision>::max(), 
			  Precision upperbound=std::numeric_limits<Precision>::max()){

  Precision meanOfSquare = 0.0;
  int count = 0;
  mean = 0.0;
  deviation = 0.0;  
  Precision max =-std::numeric_limits<Precision>::max(); 
  Precision min =std::numeric_limits<Precision>::max(); 

  for (unsigned int i=0; i<tab.size(); i++){
    if(tab[i] < min){
      min = tab[i];
      if (argmin != NULL)
	*argmin = i;
    }
    if(tab[i] > max){
      max = tab[i];
      if (argmax != NULL)
	*argmax = i;
    }
    if (tab[i] >= lowerbound &&
	tab[i] <= upperbound){
      mean += tab[i];
      meanOfSquare += tab[i] * tab[i];
      count ++;
    }
  }    
  if (count == 0){
    PRINT_ERROR("Empty set in GetMeanDev");
  }
  mean /= double(count);
  meanOfSquare /= double(count);
  minimum = min;
  maximum = max;

  deviation = sqrt(meanOfSquare - mean*mean);
}
 
template <class Precision>
inline void GetStatFromList(std::vector<Precision> tab, 
			    Precision& mean, 
			    Precision& deviation, 
			    Precision& minimum, 			    
			    Precision& Q1,
			    Precision& median,
			    Precision& Q3,
			    Precision& maximum,
			    unsigned int* argmin=NULL, 
			    unsigned int* argmax=NULL){
  if(tab.size() == 0){
    mean = deviation = minimum = Q1 = median = Q3 = maximum = 0.0;
    return;
  }
  GetMeanAndDev<Precision>(tab,mean,deviation,minimum,maximum,argmin,argmax);
  std::vector<Precision> vect(tab.begin(), tab.end());
  sort(vect.begin(), vect.end());  
  unsigned int nb = vect.size(); 
  if (nb%2 == 0){
    median = (vect[nb/2] + vect[nb/2-1])/2.0;
    unsigned int N = nb/2;
    if (N%2==0){
      Q1 = (vect[N/2] + vect[N/2-1])/2.0;
      Q3 = (vect[N+N/2] + vect[N+N/2-1])/2.0;
    }
    else{
      Q1 = vect[N/2];
      Q3 = vect[N+N/2];
    }
  }
  else {
    median = vect[nb/2];
    unsigned int N = nb/2;
    if (N%2==0){
      Q1 = (vect[N/2] + vect[N/2-1])/2.0;
      Q3 = (vect[N+1+N/2] + vect[N+1+N/2-1])/2.0;
    }
    else{
      Q1 = vect[N/2];
      Q3 = vect[N+1+N/2];
    }
  }
}


template <class T>
inline std::string StringJoin(const std::vector<T>& input, const std::string& delim){
  std::string sentence;
  for(unsigned int i=0; i<input.size(); i++){
    if (i!=0){
      sentence += delim;
    }
    sentence += to_string<T>(input[i]);
  }
  return sentence;
}

 
template <class T>
inline void StringSplit(const std::string& sentence, const std::string& delim, std::vector<T>& results)
{
  size_t cutAt;
  T val;
  results.clear();
  std::string str = sentence;
  while( (cutAt = str.find_first_of(delim)) != str.npos ){
    if(cutAt >= 0){
      std::istringstream ss(str.substr(0,cutAt));
      ss >> val;
      results.push_back(val);
    }
    str = str.substr(cutAt+1);
  }
  if(str.length() > 0){
    std::istringstream ss(str);
    ss >> val;
    results.push_back(val);
  }
}

template <class T>
inline std::vector<T> StringSplit(const std::string& sentence, const std::string& delim)
{
  std::vector<T> results;
  StringSplit<T>(sentence,delim,results);
  return results;
}

// "1,2,3;4,5,6;7,8,9" -> [[1,2,3],[4,5,6],[7,8,9]] 
// delim1 = "," , delim2 = ";"
template <class T>
inline void StringSplitToMatrix(const std::string& sentence, const std::string& delim1, const std::string& delim2, std::vector<std::vector<T> >& results)
{
  results.clear();
  std::vector<std::string> tabstr;
  StringSplit<std::string>(sentence,delim2,tabstr);
  results.resize(tabstr.size());
  for(unsigned int k = 0; k<tabstr.size(); k++){
    StringSplit<T>(tabstr[k],delim1,results[k]);  
  }
}

// ex: "1:5" -> [1,2,3,4,5]
// ex: "1:5,10,20:21" -> [1,2,3,4,5,10,20,21]
inline void StringSplitSetList(const std::string& sentence, std::list<int>& results)
{
  results.clear();
  std::vector<std::vector<int> > tabstr;
  StringSplitToMatrix(sentence,":",",",tabstr);
  for(unsigned int k = 0; k<tabstr.size(); k++){
    if (tabstr[k].size() ==2){
      for(int i = tabstr[k][0]; i <= tabstr[k][1]; ++i){
	results.push_back(i);	
      }
    }
    else{
      results.push_back(tabstr[k][0]);
    }
  }
}


// read an array file (same number of column per line)
// and write/cast value into two different matrices.
// ex file : 'filename1 x1 y1 z1'
//            filename2 x2 y2 z2'
// columnlist1 = "1"   type R: 'string'
// columnlist2 = "2:4" type S: 'double'
// =>
// mat1 = [[filename1],[filename2]]
// mat2 = [[x1,y1,z1],[x2,y2,z2]]
template <class R, class S>
void ReadSplitMatrix(std::string& filename, 
		     std::string& collist1, std::vector<std::vector<R> >& mat1, 
		     std::string& collist2, std::vector<std::vector<S> >& mat2){
  
  std::ifstream in;
  std::string buff;
  int countLine=0;
  R value1;
  S value2;
  std::string tmp;
  
  std::list<int> col1;
  std::list<int> col2;
  StringSplitSetList(collist1, col1);
  StringSplitSetList(collist2, col2);

  in.open(filename.c_str());
  if(!in) { 
    PRINT_ERROR("Unable to open the file %s",filename.c_str());
  }
  
  while (std::getline(in, buff)) {
    if (buff.size() == 0){
      continue;
    }
    std::istringstream iss(buff);
    if (buff[0] == '#'){
      //iss >> str;
      // if we read a comment, read all the line
      // do nothing
    }
    else {
      int columnCount=0;
      mat1.resize(countLine+1);
      mat2.resize(countLine+1);
      std::list<int>::const_iterator it1 = col1.begin();
      std::list<int>::const_iterator it2 = col2.begin();
      while( !iss.eof() && !iss.fail()){ 
	if (it1 != col1.end() and columnCount == (*it1)){
	  ++it1;
	  iss >> value1;
	  if (!iss.fail()){ 
	    mat1[countLine].push_back(value1);
	  }
	}
	else if (it2 != col2.end() and columnCount == (*it2)){
	  ++it2;
	  iss >> value2;
	  if (!iss.fail()){ 
	    mat2[countLine].push_back(value2);
	  }
	}
	else{
	  iss >>tmp;
	}
	columnCount++;
      }
      countLine++;
    }
  }
  in.close();
}





// matrix to string
inline std::ostream& operator<<(std::ostream& stream,std::vector<std::vector<double> >& matrix){
  for (unsigned int i=0; i<matrix.size(); i++){
    if (i!=0){
      stream << ";";  
    }
    stream << matrix[i][0];    
    for (unsigned int j=1; j<matrix[i].size(); j++){  
      stream << "," << matrix[i][j] ;
    }
  }
  return stream;
}

template<class T1, class T2>
  inline std::ostream& operator<<(std::ostream& stream, const std::vector<std::map<T1,T2> >& matrix){
  for (unsigned int i=0; i<matrix.size(); i++){
    if (i!=0){
      stream << std::endl; 
    }
    typename std::map<T1,T2>::const_iterator
      itb = matrix[i].begin(),
      ite = matrix[i].end();
    for (; itb!=ite; itb++){
      if (itb!=matrix[i].begin()){
	stream << " , ";
      }
      stream << (*itb).first<<"="<< (*itb).second ;
    }
  }
  return stream;
}


// convert a string of format "1:2, 3:4, 5:6" into a MAP
// delim1 = "," 
// delim2 = ":"
template <class T1,class T2>
  inline void StringSplitIntoMap(const std::string& sentence, const std::string& delim1, const std::string& delim2, std::map<T1,T2>& results)
{
  size_t cutAt, cutAt2;
  T1 val1;
  T2 val2;
  results.clear();
  std::string str = sentence;
  while( (cutAt = str.find_first_of(delim1)) != str.npos ){
    if(cutAt > 0){
      std::string subpart = str.substr(0,cutAt);
      if( (cutAt2 = subpart.find_first_of(delim2)) != str.npos ){
	std::istringstream ss1(subpart.substr(0,cutAt2));
	ss1 >> val1;
	std::istringstream ss2(subpart.substr(cutAt2+1,cutAt-cutAt2-1));
	ss2 >> val2;
      }
      results[val1] = val2;
    }
    str = str.substr(cutAt+1);
  }
  if(str.length() > 0){
    if( (cutAt2 = str.find_first_of(delim2)) != str.npos ){
      std::istringstream ss1(str.substr(0,cutAt2));
      ss1 >> val1;
      std::istringstream ss2(str.substr(cutAt2+1,str.length()-cutAt2-1));
      ss2 >> val2;
    }
    results[val1] = val2;
  }
}



// extract a list of name store in the key KEY of mapstring MAP
// and retreive the corresponding feature of type T in the structure LIST
template <class T>
inline void extract_list(const std::map<std::string, std::string>& map,
			 std::string& key,
			 const std::map<std::string, T>& list,
			 std::vector<T>& output)
{
  std::vector<std::string> itemsStr;
  // Read "toto1,toto2,toto3,..." and split into a vector of string
  StringSplit<std::string>(map_extract<std::string>(map, key),
			   ",", itemsStr);
  // add the list["toto1"], list["toto2"], ... in the output vector
  for(unsigned int i =0; i<itemsStr.size(); i++){
    output.push_back(map_extract<T>(list, itemsStr[i]));
  }
  
}


// in order to seed a random generator we need a integer
// this function generate an integer from a string
inline int StringToInt(const std::string& word){
  int i=1;
  
  std::string::const_iterator itb = word.begin(), ite= word.end();
  for (; itb!=ite; itb++){
    i *= (int)(*itb);
  }
  return i;
}


// replace function like in python
inline std::string Replace(const std::string& word, const std::string& query, const std::string replace){
  int end= word.rfind(query);
  std::string newStr = word;
  newStr.replace(end,query.size(), replace);
  return newStr;
}



// Read a file return the vector of first element of each line
inline void GetStringList(const char * filename, 
			  std::vector<std::string>& list){
  std::ifstream in;
  std::string buff;  
  in.open(filename);
  if(!in) {
    PRINT_ERROR("Unable to open the file.");
  }
  in >> buff;
  while( !in.eof() ){
    if (buff[0] == '#'){
      // if we read a comment, read all the line
      getline(in, buff);
    }
    else {
      list.push_back(buff);
      getline(in, buff);
    }
    in >> buff;
  }
  
  in.close();
}






/*
  parseCommandLine read the string table argv and return the list of args 
  and the list of option
*/
inline void ParseCommandLine(int argc, char **argv,
			     std::string format,
			     std::vector<std::string>& args,
			     std::vector<std::pair<std::string,std::string> >& opt)
{

  int index = 1;
  while (index < argc){
    std::string word(argv[index]);
    if (word[0] == '-'  && word.size()>1){
      char c = word[1]; 
      if (c == '-'){
	PRINT_ERROR("Long option not yet implemented");
      }
      size_t rank = format.find(c);
      if (rank == std::string::npos){
	PRINT_ERROR("The option '%s' is not expected, verify getopt pattern.",word.c_str());
      }
      else if(rank+1 < format.length() && format[rank+1]==':'){
	opt.push_back(std::pair<std::string,std::string>(std::string(word), std::string(argv[index+1])));
	index++;      
      }
      else{
	opt.push_back(std::pair<std::string,std::string>(std::string(word), std::string("")));
      }
    }
    else{
      args.push_back(word);
    }
    index++;      
  } 
}


//cc UNIX path only, use library instead
inline std::string Basename(const std::string& name){
  std::vector<std::string> tokens;  
  StringSplit(name, "/", tokens);
  return tokens.back();
}

// get dir (UNIX only)
inline std::string Dir(const std::string& name){
  std::vector<std::string> tokens;  
  StringSplit(name, "/", tokens);
  tokens.pop_back();
  return StringJoin<std::string>(tokens,"/");
}


// erase last extension of the string
inline std::string NoExtension(const std::string& name){
  std::vector<std::string> tokens;  
  StringSplit(name, ".", tokens);
  tokens.pop_back();
  return StringJoin<std::string>(tokens,".");
}

inline std::string Extension(const std::string& name){
  std::vector<std::string> tokens;  
  StringSplit(name, ".", tokens);
  return  tokens.back();
}

template <class T>
inline std::vector<T> ExtractKeys(const std::string key, const std::vector<std::map<std::string, T> >& vectmap){
  std::vector<T> keys;
  for(unsigned int i = 0; i<vectmap.size(); i++){
    keys.push_back(map_extract(vectmap[i], key ));
  }
  return keys;
}





class Combination
{
 public:
  
  static unsigned long long Factorial(int n){
    unsigned long long res = 1;
    for (int i=1; i<=n; i++){
      res *= i;
    }
    return res;
  }
  
  static unsigned int BinomialCoef(int n, int k){
    if (n<k){
      PRINT_ERROR("Wrong input");
    }
    // k is probably very small usually 2 or 3.
    unsigned int N_Fact_On_NminusK_Fact = 1;
    unsigned int K_Fact = 1;
    for (int i=n, j=1; i>(n-k); i--, j++){
      N_Fact_On_NminusK_Fact *= i;
      K_Fact *= j;
    }
    return N_Fact_On_NminusK_Fact/K_Fact;
  }


  // This function compute the next size-long code 
  // where each element is inferior to n and where repetition are allowed
  // ex with n=4 and size=3:
  // 0,0,0  -> 1,0,0
  // 1,0,0  -> 2,0,0
  // 2,0,0  -> 3,0,0
  // 3,0,0  -> 0,1,0
  // 0,1,0  -> 1,1,0
  // ...
  static inline void NextCode(int n, int size, std::vector<int>& combi){
    int remain = 1;
    for (int i=0; i<size; i++){
      int v = combi[i]+remain;
      // Here a trivial solution would have been to use 
      // a modulo/division which is more that what we need
      // and is more computational.
      // we only want the remainder to be 1 when v-n =0
      // and we want combi[i] equal to zero if the remainder is 1.
      remain = (int)(v-n ==0);
      combi[i] = ~((!remain)-1) & v;
    }
  }
  
  

  // This function compute the next size-long combination 
  // where each element is inferior to n and where repetition are not allowed
  // The combination is always ordered (smaller to bigger int)
  // ex with n=4 and size=3:
  // 0,1,2  -> 0,1,3
  // 0,1,3  -> 0,2,3
  // 0,2,3  -> 1,2,3
  static inline void NextCombi(int n, int size, std::vector<int>& combi){
    // verification end reached
    //if (combi[0] == n - size)
    //  return ;

    int i = size - 1; 
    while(i > 0 && combi[i] == n - size + i){
      --i;
    }
    
    ++combi[i];
    
    for (int j = i; j < size - 1; ++j){
      combi[j+1] = combi[j] + 1;
    }
  }
  
  // Given a target list of integer T of size N 
  // and a correponding vector of list of candidate V
  // Determine if there exist a vector of integer S and an offset such as
  // for all i, V[(i+offset)%N)][S[i]] = T[i]
  static inline 
    bool IsThereAToricCorrespondance(std::vector<int>& T,
				     std::vector<std::set<int> >& V){
    unsigned int N = T.size();
    for (unsigned int offset=0; offset<N ; offset++){
      unsigned int validNb = 0;
      for(unsigned int i=0; i<N; i++){	
	int exist = 0;
	std::set<int>::const_iterator
	  cand     = (V[(i+offset)%N]).begin(),
	  cand_end = (V[(i+offset)%N]).end();
	for(; cand!=cand_end; cand++){
	  exist += (int)((*cand)==T[i]);
	}
	validNb+=(int)(exist!=0);
      }
      if (validNb == N){
	return true;
      }
    }
    return false;
  }
  

  // Given a target list of integer T of size N 
  // and a correponding vector of list of candidate V
  // return if there is a vector of integer S such as
  // all V[i][S[i]] are different and belong to T 
  // Doesnt work if T and V have different size
  // NO LONGER WORK 
  static inline bool IsThereACorrespondance(std::vector<int>& T,
					    std::vector<std::set<int> >& V){
    unsigned int N = T.size();
    float a[N][N];
    for(unsigned int i=0; i<N; i++){// on T	
      for(unsigned int j=0; j<N; j++){// on V
	std::set<int>& candSet= (V[j]);
	a[i][j] = (int)(candSet.find(T[i]) != candSet.end())*(i+j*N)*(i+j*N);
      }
    }
    // KIND OF DETERMINANT USING INTEGER ONLY
    float mult;
    float deter=1;
    for(unsigned int i=0;i<N;i++){
      for(unsigned int j=0;j<N;j++){
    	mult=(a[i][i]==0)?a[j][i] : a[j][i]/a[i][i];
    	for(unsigned int k=0;k<N;k++){
    	  if(i==j) break;
    	  a[j][k] += -a[i][k]*mult;
    	}
      }
    }
    for(unsigned int i=0;i<N;i++){
      deter *= a[i][i];
    }
   
    return (deter!=0);
  }


  static inline int Intersect(std::vector<int>& T,
			      std::set<int>& V){
    int rep =0;
    for(unsigned int i=0; i<T.size(); i++){
      rep += (int)(V.find(T[i]) != V.end());	
    }
    return (rep!=0);
  }


  static inline int MissingNb(std::vector<int>& T,
			      std::vector<std::set<int>* >& V){
    unsigned int N = T.size();
    int missingNb=0;
    for(unsigned int i=0; i<N; i++){// on T	
      int a=0;
      for(unsigned int j=0; j<N; j++){// on V
	std::set<int>& candSet= (*V[j]);
	a += (int)(candSet.find(T[i]) != candSet.end());
      }
      missingNb += (int)(a==0);
    }
    return missingNb;
  }


  //#include <cmath>
  //static inline std::vector<int> Permutation(int k, int size){
  //  int j=1;
  //  int p,i;
  //  std::vector<int> r(size,0);
  //  r[size-1] = 0;
  //  while(j++<size){
  //    p = i = size-j;
  //    r[p] = floor((k/=(j-1))%j);
  //    while(i++<size)
  //	if(r[i] >= r[p])
  //	  ++r[i];
  //  }
  //  return r;
  //}

};

#endif
