//
// Copyright (C) 2010  Clement Creusot
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef STOPWATCHH
#define STOPWATCHH

#include <list>
#include <time.h>
#include <algorithm> // for find_if()
#include <string>
#include <cstdio>

// WARNING: This classes are not thread safe

// WARNING: If you link your code to any file using a stopwatch
// functionnality, you have to call STOPWATCH_INIT in your main file.

// Initialisation of the static variables
// Call it before main.
#define STOPWATCH_INIT()			\
  clock_t Stopwatch::programStart;		\
  clock_t Stopwatch::programEnd;		\
  std::list<TimeEntry> Stopwatch::chronoList;	\
  int Stopwatch::level;	

static inline void Seconds2HourMinuteSecond(int secondsIn, int& hours, int& minutes, int& seconds){
  int total= secondsIn;
  hours = total/3600;
  total = total - hours*3600;
  minutes = total/60;
  total = total - minutes*60;
  seconds = total;
}

 
class TimeEntry
{
 public: 

  TimeEntry(const std::string& label){
    this->name = label;
    this->isStatusOn = false;
    this->totalTime = 0.0;
    this->level = -1; //not yet determined
  }
  
  void Start(int lvl=0){
    if (this->isStatusOn){
      fprintf(stderr, "WARNING: Call start on an running stopwatch.\n");
      this->Stop();
    }
    currentStart = clock();
    this->isStatusOn = true;
    if (this->level == -1){
      this->level = lvl;
    }
    else if (this->level != lvl){
      this->level = -2;
    }
    else{
      this->level = lvl;
    }
  }

  void Stop(){
    if (this->isStatusOn){
      this->isStatusOn = false;
      currentStop = clock();
      double diff = (currentStop-currentStart)/(double)CLOCKS_PER_SEC;
      this->periods.push_back(diff);
      this->totalTime += diff;
    }
    else
      fprintf(stderr, "WARNING: Call stop on non-running stopwatch.\n");
  }
  
  void GetData(double& time, int& count, double& mean, int& level){
    if (this->isStatusOn){
      fprintf(stderr, "WARNING: Call GetData on an running stopwatch.\n");
      this->Stop();
    }
    time = this->totalTime;
    count = this->periods.size();
    mean = 0.0;
 
    std::list<double>::iterator itb = this->periods.begin(),
      ite = this->periods.end();
    for (; itb!=ite; itb++){
      mean += (*itb);
    }    
    mean /= count;
    level = this->level;
  }

  const std::string& GetName() const{return this->name;}

 private:
  std::string name;
  int level;
  bool isStatusOn;
  std::list<double> periods;
  clock_t currentStart;
  clock_t currentStop;
  double totalTime;
};


// In order to find a TimeEntry using only the label
struct TimeEntry_Finder
{
TimeEntry_Finder(const std::string& label )
: target_label(label){
};
  bool operator()( TimeEntry const& test) const{
    return test.GetName().compare(target_label) == 0;
  };
  const std::string& target_label;
};




class Stopwatch
{
 private:
  // No public constructor for this class
  // only use static method
  Stopwatch () {level = 0;};
  
  static clock_t programStart;
  static clock_t programEnd;
  static std::list<TimeEntry> chronoList;
  static int level;

 public: 

  static void ProgramStart(){
    programStart = clock();
  }
  
  static void ProgramStop(){
    programEnd = clock();
  }
  
  static void Start(const std::string label){
    std::list<TimeEntry>::iterator cur;
    cur = std::find_if(chronoList.begin(), 
		       chronoList.end(), 
		       TimeEntry_Finder(label));
    if (cur == chronoList.end() ){
      chronoList.push_back(TimeEntry(label));
      chronoList.back().Start(level);
    }
    else{
      (*cur).Start(level);
    }
    level++;
  }


  static void Stop(const std::string label){
    std::list<TimeEntry>::iterator cur;
    cur = std::find_if(chronoList.begin(), 
		       chronoList.end(), 
		       TimeEntry_Finder(label));
    if (cur == chronoList.end() ){
      fprintf(stderr, "WARNING: Call Stop on unexistant TimeEntry.\n");
    }
    else{
      (*cur).Stop();
      level--;
    } 
  }

  
  static void PrintDetails(){
    double total, curTotal, mean;
    int count, curlevel;
    total = (programEnd-programStart)/ (double)CLOCKS_PER_SEC;  
    
    fprintf(stderr,"======================================================================\n");
    fprintf(stderr,"== Time Details ==\n");
    fprintf(stderr,"==================\n");
    fprintf(stderr,"  Total Time: %f\n\n", total);
    
    std::list<TimeEntry>::iterator 
      itb = chronoList.begin(), 
      ite = chronoList.end(); 

    fprintf(stderr,"  %-27s|", "name");
    fprintf(stderr,"  %-7s|", "total");
    fprintf(stderr,"  %-7s|", "count");
    fprintf(stderr,"  %-7s|", "mean");
    fprintf(stderr,"  %-7s|\n", "prop");
    fprintf(stderr,"----------------------------------------------------------------------\n");
    for (; itb!=ite; itb++){
      (*itb).GetData(curTotal, count, mean, curlevel);
      std::string name((*itb).GetName());
      if (curlevel >0){
	name.insert(0,curlevel,'.');
      }
      fprintf(stderr,"  %-26s |", name.substr(0,26).c_str());
      fprintf(stderr,"  %6.2f |", curTotal);      
      fprintf(stderr,"  %6d |", count);      
      fprintf(stderr,"  %6.2f |", mean);      
      fprintf(stderr,"  %5.1f%% |\n", 100*curTotal/total);      
    }

    fprintf(stderr,"======================================================================\n");

  }

};


#endif

