//
// Copyright (C) 2010  Clement Creusot
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "stopwatch.h"
#include "common.h"

#include "geom.h"
#include "byProduct.h"

ByProduct::ByProduct(std::vector<double>& f)
  : field(f)
{
}

ByProduct::~ByProduct()
{  
}




void ByProduct::Compute(std::string method, 
			std::vector<std::vector<double>*>& items,
			const std::map<std::string,std::string>& prop)
{  
  if (method.compare("mean") ==0){
    if (items.size() < 2){
      PRINT_ERROR("Wrong nb of dependances for the %s method.", method.c_str());
    }
    ComputeMean(items);
  }
  else if (method.compare("max") ==0){
    if (items.size() < 2){
      PRINT_ERROR("Wrong nb of dependances for the %s method.", method.c_str());
    }
    ComputeMax(items);
  }
  else if (method.compare("normalize") ==0){
    if (items.size() < 1){
      PRINT_ERROR("Wrong nb of dependances for the %s method.", method.c_str());
    }
    ComputeNormalize(*items[0]);
  }
  else if (method.compare("maxN") ==0){
    if (items.size() < 2){
      PRINT_ERROR("Wrong nb of dependances for the %s method.", method.c_str());
    }
    ComputeMaxNormalized(items);
  }
  else if (method.compare("weightedSum") ==0){
    if (items.size() < 2){
      PRINT_ERROR("Wrong nb of dependances for the %s method.", method.c_str());
    }
    std::vector<double> weights;
    StringSplit<double>(map_extract(prop,"weights"),",",weights);
    ComputeWeightedSum(items, weights);
  }
  else if (method.compare("weightedSumN") ==0){
    if (items.size() < 2){
      PRINT_ERROR("Wrong nb of dependances for the %s method.", method.c_str());
    }
    std::vector<double> weights;
    StringSplit<double>(map_extract(prop,"weights"),",",weights);
    ComputeWeightedSumNormalized(items, weights);
  }
  else if (method.compare("product") ==0){
    if (items.size() != 2){
      PRINT_ERROR("Wrong nb of dependances for the %s method.", method.c_str());
    }
    ComputeProduct((*items[0]), (*items[1]));    
  }
  else if (method.compare("shapeIndex") ==0){
    if (items.size() != 2){
      PRINT_ERROR("Wrong nb of dependances for the %s method.", method.c_str());
    }
    ComputeShapeIndex((*items[0]), (*items[1]));    
  }

  else if (method.compare("curvedness") ==0){
    if (items.size() != 2){
      PRINT_ERROR("Wrong nb of dependances for the %s method.", method.c_str());
    }
    ComputeCurvedness((*items[0]), (*items[1]));    
  }
  else if (method.compare("logCurvedness") ==0){
    if (items.size() != 2){
      PRINT_ERROR("Wrong nb of dependances for the %s method.", method.c_str());
    }
    ComputeLogCurvedness((*items[0]), (*items[1]));    
  }
  else if (method.compare("willmore") ==0){
    if (items.size() != 2){
      PRINT_ERROR("Wrong nb of dependances for the %s method.", method.c_str());
    }
    ComputeWillmore((*items[0]), (*items[1]));    
  }
  else if (method.compare("dibeklioglu") ==0){
    if (items.size() != 2){
      PRINT_ERROR("Wrong nb of dependances for the %s method.", method.c_str());
    }
    ComputeDibeklioglu((*items[0]), (*items[1]));    
  }
  else if (method.compare("diff") ==0){
    if (items.size() != 2){
      PRINT_ERROR("Wrong nb of dependances for the %s method.", method.c_str());
    }
    ComputeDiff((*items[0]), (*items[1]));    
  }
  else{
    PRINT_ERROR("Unknown by-product method %s.",method.c_str());
  }
}


void ByProduct::ComputeMean(std::vector<std::vector<double>*>& items){
  unsigned int size = (*items[0]).size();
  for (unsigned int i=0; i<items.size(); i++){
    if ( (*items[i]).size() != size ){
      PRINT_ERROR("The fields dont have the same size.");
    }
  }
  this->field.resize(size,0.0);
  for (unsigned int i=0; i<items.size(); i++){
    for(unsigned int k = 0; k<size; k++){
      this->field[k] += (*items[i])[k];
    } 
  }
  double coef = 1/(double)items.size();
  for(unsigned int k = 0; k<size; k++){
    this->field[k] *= coef;
  } 
}

void ByProduct::ComputeMax(std::vector<std::vector<double>*>& items){
  unsigned int size = (*items[0]).size();
  for (unsigned int i=0; i<items.size(); i++){
    if ( (*items[i]).size() != size ){
      PRINT_ERROR("The fields dont have the same size.");
    }
  }
  this->field.resize(size,-std::numeric_limits<double>::max());
  for (unsigned int i=0; i<items.size(); i++){
    for(unsigned int k = 0; k<size; k++){
      this->field[k] = std::max(this->field[k],(*items[i])[k]);
    } 
  }
}

void ByProduct::ComputeMaxNormalized(std::vector<std::vector<double>*>& items){
  unsigned int size = (*items[0]).size();
  std::vector<double> mins(items.size(),std::numeric_limits<double>::max());
  std::vector<double> maxs(items.size(),-std::numeric_limits<double>::max());
  for (unsigned int i=0; i<items.size(); i++){
    if ( (*items[i]).size() != size ){
      PRINT_ERROR("The fields dont have the same size.");
    }
    for(unsigned int k = 0; k<size; k++){
      mins[i] = std::min(mins[i],(*items[i])[k]);
      maxs[i] = std::max(maxs[i],(*items[i])[k]);
    }    
  }
  this->field.resize(size,-std::numeric_limits<double>::max());
  for (unsigned int i=0; i<items.size(); i++){
    double domain = maxs[i] - mins[i];
    for(unsigned int k = 0; k<size; k++){
      double value = ((*items[i])[k]-mins[i])/domain;
      this->field[k] = std::max(this->field[k],value);
    } 
  }
}

void ByProduct::ComputeWeightedSum(std::vector<std::vector<double>*>& items,
				   std::vector<double>& weights){
  if (weights.size() != items.size()){
    PRINT_ERROR("Items and weights dont match.");
  }
  unsigned int size = (*items[0]).size();
  for (unsigned int i=0; i<items.size(); i++){
    if ( (*items[i]).size() != size ){
      PRINT_ERROR("The fields dont have the same size.");
    }
  }
  this->field.resize(size,0);
  for(unsigned int k = 0; k<size; k++){
    double value = 0.0;
    for (unsigned int i=0; i<items.size(); i++){
      value += weights[i]* (*items[i])[k];
    } 
    this->field[k]  = value;
  }
}

void ByProduct::ComputeWeightedSumNormalized(std::vector<std::vector<double>*>& items,
					     std::vector<double>& weights){
  if (weights.size() != items.size()){
    PRINT_ERROR("Items and weights dont match.");
  }
  unsigned int size = (*items[0]).size();
  std::vector<double> mins(items.size(),std::numeric_limits<double>::max());
  std::vector<double> maxs(items.size(),-std::numeric_limits<double>::max());

  for (unsigned int i=0; i<items.size(); i++){
    if ( (*items[i]).size() != size ){
      PRINT_ERROR("The fields dont have the same size.");
    }
    for(unsigned int k = 0; k<size; k++){
      mins[i] = std::min(mins[i],(*items[i])[k]);
      maxs[i] = std::max(maxs[i],(*items[i])[k]);
    }    
  }

  this->field.resize(size,0.0);
  for (unsigned int i=0; i<items.size(); i++){
    double coef = 1/(maxs[i] - mins[i]);
    for(unsigned int k = 0; k<size; k++){
      double value = ((*items[i])[k]-mins[i])*coef;
      this->field[k] +=  weights[i]* value;
    } 
  }
}




void ByProduct::ComputeWeightedSumWeightNormalised(std::vector<std::vector<double>*>& items,
						   std::vector<double>& weights){
  if (weights.size() != items.size()){
    PRINT_ERROR("Items and weights dont match.");
  }
  unsigned int size = (*items[0]).size();
  double mini = 0.0;
  double maxi = 0.0;
  for (unsigned int i=0; i<items.size(); i++){
    if ( (*items[i]).size() != size ){
      PRINT_ERROR("The fields dont have the same size.");
    }
    if (weights[i] > 0.0){
      maxi += weights[i];
    }
    else{
      mini += weights[i];
    }
  }
  if (maxi-mini == 0.0){
    PRINT_ERROR("Weights are null");
  }
  double ratio = 1/(maxi - mini);
  std::cerr << mini << " " << maxi << std::endl;
  this->field.resize(size,0.0);
  for(unsigned int k = 0; k<size; k++){
    double value = 0.0;
    for (unsigned int i=0; i<items.size(); i++){
      value += (*items[i])[k]*weights[i];
    } 
    this->field[k] = (value-mini)*ratio;
  }
}

void ByProduct::ComputeNormalize(std::vector<double>& vect){
  unsigned int size = vect.size();
  this->field.resize(size,0.0);
  double mini = std::numeric_limits<double>::max();
  double maxi = -std::numeric_limits<double>::max();
  for (unsigned int i=0; i<vect.size(); i++){
    mini = std::min(mini,vect[i]);
    maxi = std::max(maxi,vect[i]);
  }
  double domain = maxi - mini;
  double nmini = std::numeric_limits<double>::max();
  double nmaxi = -std::numeric_limits<double>::max();

  for(unsigned int k = 0; k<size; k++){
    this->field[k] = (vect[k]-mini)/domain ;
    nmini = std::min(nmini,field[k]);
    nmaxi = std::max(nmaxi,field[k]);
  }
  std::cerr << "MINMAX " << mini << " " << maxi << " " << nmini << " " << nmaxi <<  std::endl;
  
}


void ByProduct::ComputeProduct(std::vector<double>& item1, 
			       std::vector<double>& item2){
  if (item1.size() != item2.size()){
    PRINT_ERROR("The fields dont have the same size.");
  }
  unsigned int size = item1.size(); 
  this->field.resize(size,0.0);
  for(unsigned int i = 0; i<size; i++){
    this->field[i] = item1[i] * item2[i];
  } 
}


void ByProduct::ComputeShapeIndex(std::vector<double>& item1, 
				  std::vector<double>& item2){
  if (item1.size() != item2.size()){
    PRINT_ERROR("The fields dont have the same size.");
  }
  unsigned int size = item1.size(); 
  this->field.resize(size,0.0);
  for(unsigned int i = 0; i<size; i++){
    this->field[i] = 1/2.0 - 1/PI * atan((item1[i] + item2[i])/(item1[i] - item2[i]));
  } 
}

void ByProduct::ComputeCurvedness(std::vector<double>& item1, 
				  std::vector<double>& item2){
  if (item1.size() != item2.size()){
    PRINT_ERROR("The fields dont have the same size.");
  }
  unsigned int size = item1.size(); 
  this->field.resize(size,0.0);
  for(unsigned int i = 0; i<size; i++){
    this->field[i] = sqrt((item1[i]*item1[i]+item2[i]*item2[i])/2);
  } 
}


void ByProduct::ComputeLogCurvedness(std::vector<double>& item1, 
				     std::vector<double>& item2){
  if (item1.size() != item2.size()){
    PRINT_ERROR("The fields dont have the same size.");
  }
  unsigned int size = item1.size(); 
  this->field.resize(size,0.0);
  for(unsigned int i = 0; i<size; i++){
    this->field[i] = 2/PI * log(sqrt((item1[i]*item1[i]+item2[i]*item2[i])/2));
  } 
}


// take k1 and k2 in input
// DONT USE THIS : use discrete willmore energy instead
void ByProduct::ComputeWillmore(std::vector<double>& item1, 
				std::vector<double>& item2){
  if (item1.size() != item2.size()){
    PRINT_ERROR("The fields dont have the same size.");
  }
  unsigned int size = item1.size(); 
  this->field.resize(size,0.0);
  for(unsigned int i = 0; i<size; i++){
    this->field[i] = 1/double(4) * (item1[i]-item2[i]) * (item1[i]-item2[i]);
  } 
}


// take k1 and k2 in input
// compute ln(H-K+1) as in [Dibeklioglu2008] 
void ByProduct::ComputeDibeklioglu(std::vector<double>& item1, 
				   std::vector<double>& item2){
  if (item1.size() != item2.size()){
    PRINT_ERROR("The fields dont have the same size.");
  }
  unsigned int size = item1.size(); 
  this->field.resize(size,0.0);
  for(unsigned int i = 0; i<size; i++){
    this->field[i] = log((item1[i]*item2[i]) - 0.5*(item1[i]+item2[i]) +1);
  } 
}


void ByProduct::ComputeDiff(std::vector<double>& item1, 
			    std::vector<double>& item2){
  if (item1.size() != item2.size()){
    PRINT_ERROR("The fields dont have the same size.");
  }
  unsigned int size = item1.size(); 
  this->field.resize(size,0.0);
  for(unsigned int i = 0; i<size; i++){
    this->field[i] = item1[i]-item2[i];
  } 
}
