//
// Copyright (C) 2010  Clement Creusot
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include"goldfeatherCubicSimple.h"

#include"../geom.h"
#include"common.h"
#include"stopwatch.h"
#include <armadillo>
using namespace arma;


colvec Solve(mat& A, colvec& b){
  Stopwatch::Start("solve");
  colvec tmp = solve(A,b);
  Stopwatch::Stop("solve");  
  return tmp;
}


colvec SolveLIC(mat& A, colvec& b){

  Stopwatch::Start("LIC");
  colvec tmp = inv(trans(A)*A)*trans(A)*b ; // linearly independant columns
  Stopwatch::Stop("LIC");
  return tmp;
}



// normal-enhanced cubic fitting: Goldfeather and Interrante TOG 23(1)
void GoldfeatherCubicSimple::Compute(const std::vector<int>* subset){

  Stopwatch::Start("curvature");
  Stopwatch::Start("normal");
  this->mesh->MakeNormals(this->normals);
  Stopwatch::Stop("normal");
  
  colvec (*SolveFunc_)(mat&, colvec&) = NULL; 
  if (method.compare("LIC")==0){
    SolveFunc_ = &SolveLIC;
  }
  //else if(method.compare("SVD")==0){
  //  SolveFunc_ = &SolveSVD;
  //}
  //else if (method.compare("QR")==0){
  //  SolveFunc_ = &SolveQR;
  //}
  //else if (method.compare("LIR")==0){
  //  SolveFunc_ = &SolveLIR;
  //}
  else if (method.compare("generic")==0){
    SolveFunc_ = &Solve;
  }
  else {
    SolveFunc_ = &SolveLIC;
  }
  
  bool useSubset = (subset!=NULL);
  int meshSize = (int)this->mesh->Points().size();
  
  std::vector<int>::const_iterator itb,ite(itb);
  if (useSubset){
    itb = (*subset).begin();
    ite = (*subset).end();
  }
  for(int k=0; (useSubset && (itb!=ite)) || (!useSubset && k<meshSize) ; itb++, k++){
    int i = (useSubset) ? (*itb) : k ;  
    int size = (int)this->neighborhood->GetNeighborhoodPoints()[i].size();

    // no neighborhood, nothing to compute
    if (size == 0){
      ks1[i] = 0;
      ks2[i] = 0;
      k1[i] = 0;
      k2[i] = 0;
      continue;
    }
    
    // 3 equations per neighborhing points:
    //   - 1 for the point
    //   - 2 for its notmal
    // 7 unknown variables
    mat A(3*size,7);
    colvec b(3*size);
    

    double dirU[3] = {0.0,0.0,0.0};
    double dirV[3] = {0.0,0.0,0.0};
  
    // create a orthonormalized base from the normal
    Geom<double>::setVirtualTangents(this->normals[i],dirU,dirV);

    std::vector<int>::const_iterator 
      itnb = this->neighborhood->GetNeighborhoodPoints()[i].begin(),
      itne = this->neighborhood->GetNeighborhoodPoints()[i].end();
    for(int j=0; itnb!=itne ; j++,itnb++){

      double vect1[3];
      //make vector linking i and the current vertex
      Geom<double>::Substract_3(vect1, this->mesh->Points()[*itnb], this->mesh->Points()[i]);
     
      // project on the base (Normal,dirU,dirV)
      double zi = Geom<double>::DotProduct_3(vect1,this->normals[i]);
      double xi = Geom<double>::DotProduct_3(vect1,dirU);
      double yi = Geom<double>::DotProduct_3(vect1,dirV);      
      double x2i = xi*xi;
      double xyi = xi*yi;
      double y2i = yi*yi;
      
      // add equation for each neighboring point

      A(3*j,0) = 0.5*x2i;
      A(3*j,1) = xyi;
      A(3*j,2) = 0.5*y2i;
      A(3*j,3) = xi*x2i;
      A(3*j,4) = x2i*yi;
      A(3*j,5) = xi*y2i;
      A(3*j,6) = y2i*yi;
	       
      b(3*j) = zi;     
      
      // add equations for corresponding normal

      A(3*j+1,0) = xi;
      A(3*j+1,1) = yi;
      A(3*j+1,2) = 0.0;
      A(3*j+1,3) = 3.0*x2i;
      A(3*j+1,4) = 2.0*xyi;
      A(3*j+1,5) = y2i;
      A(3*j+1,6) = 0.0;
             
      A(3*j+2,0) = 0.0;
      A(3*j+2,1) = xi;
      A(3*j+2,2) = yi;
      A(3*j+2,3) = 0.0;
      A(3*j+2,4) = x2i;
      A(3*j+2,5) = 2.0*xyi;
      A(3*j+2,6) = 3.0*y2i;


      double ai = Geom<double>::DotProduct_3(this->normals[*itnb],dirU);
      double bi = Geom<double>::DotProduct_3(this->normals[*itnb],dirV);
      double ci = Geom<double>::DotProduct_3(this->normals[*itnb],this->normals[i]);
      b(3*j+1) = -ai/ci;
      b(3*j+2) = -bi/ci;

    }



    
    colvec x = SolveFunc_(A,b);
    double ApC = x(0)+x(2);
    double sqrtdelta = sqrt(ApC*ApC-4*(x(0)*x(2)-x(1)*x(1)));
    k1[i] = 0.5*(ApC + sqrtdelta);
    k2[i] = 0.5*(ApC - sqrtdelta);

    //if(eid[1]<eid[2]){
    //  k1[i] = eid[2];
    //  k2[i] = eid[1];
    //  t2[i][0] = (dirU[0])*eiv[1][1] + (dirV[0])*eiv[2][1];
    //  t2[i][1] = (dirU[1])*eiv[1][1] + (dirV[1])*eiv[2][1];
    //  t2[i][2] = (dirU[2])*eiv[1][1] + (dirV[2])*eiv[2][1];
    //  t1[i][0] = (dirU[0])*eiv[1][2] + (dirV[0])*eiv[2][2];
    //  t1[i][1] = (dirU[1])*eiv[1][2] + (dirV[1])*eiv[2][2];
    //  t1[i][2] = (dirU[2])*eiv[1][2] + (dirV[2])*eiv[2][2];
    //  
    //  t1x = eiv[1][2];
    //  t1y = eiv[2][2];
    //  t2x = eiv[1][1];
    //  t2y = eiv[2][1];
    //}else{
    //  k1[i] = eid[1];
    //  k2[i] = eid[2];
    //  
    //  t1[i][0] = (dirU[0])*eiv[1][1] + (dirV[0])*eiv[2][1];
    //  t1[i][1] = (dirU[1])*eiv[1][1] + (dirV[1])*eiv[2][1];
    //  t1[i][2] = (dirU[2])*eiv[1][1] + (dirV[2])*eiv[2][1];
    //  t2[i][0] = (dirU[0])*eiv[1][2] + (dirV[0])*eiv[2][2];
    //  t2[i][1] = (dirU[1])*eiv[1][2] + (dirV[1])*eiv[2][2];
    //  t2[i][2] = (dirU[2])*eiv[1][2] + (dirV[2])*eiv[2][2];
    //  t2x = eiv[1][2];
    //  t2y = eiv[2][2];
    //  t1x = eiv[1][1];
    //  t1y = eiv[2][1]; 
    //}
    //  Geom<double>::Normalize_3(t1[i]);
    //  Geom<double>::Normalize_3(t2[i]);
      
    // computing first derivative of principal curvature w.r.t their direction
    //ks1[i] = getk1t1(t1x,t1y);
    //ks2[i] = getk1t1(t2x,t2y);
  }

  Stopwatch::Stop("curvature");
}



