//
// Copyright (C) 2010  Clement Creusot
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "stopwatch.h"
#include "LSD.h"
#include "common.h"

#include "../geom.h"

LocalSymetryDescriptor::LocalSymetryDescriptor(const Neighborhood * n ,
					       double radiusBinSize,
					       double angleBinNb,
					       const std::vector<Histogram*>* normals,
					       const std::vector<Histogram*>* startingDirections,
					       const std::vector<double>* values)
{
  this->neighborhood = n;
  this->normals = normals;
  this->startingDirections = startingDirections;
  this->values = values;
  this->radiusBinSize = radiusBinSize;
  this->angleBinNb =  angleBinNb;
  Stopwatch::Start("LSD");
  this->Compute();
  Stopwatch::Stop("LSD");

}

LocalSymetryDescriptor::~LocalSymetryDescriptor()
{
  
}

void LocalSymetryDescriptor::Compute()
{  
  const Mesh * mesh = this->neighborhood->GetMesh();
  unsigned int npoint = mesh->Points().size();
  
  this->LSD.resize(npoint);

  double radiusStepSize(this->radiusBinSize); 
  int angleBinNb(this->angleBinNb);
  double angleStepSize = 2*PI/double(angleBinNb); 
  if (this->neighborhood->GetType() !=EUCLIDIAN_DISTANCE ){
    PRINT_ERROR("unsupported neighborhood for Local Symetry Descriptor");
  }
  int radiusBinNb = this->neighborhood->GetDistance()/radiusStepSize +1;


  double X[] = {1.0,0.0,0.0};
 
  for(unsigned int i=0; i<npoint; i++){

    PolarHistogram histogram;
    histogram.SetSize(radiusBinNb,angleBinNb,0.0);
    HMatrix histocount;
    histocount.SetSize(radiusBinNb,angleBinNb,0.0);

    double * N = dynamic_cast<HVector*>((*this->normals)[i])->GetRawData();
    double * Iaux = X;
    if (this->startingDirections != NULL){
      Iaux = dynamic_cast<HVector*>((*this->startingDirections)[i])->GetRawData();
    }
    double I[3], J[3];
    //Create N,I,J orthonormal base
    Geom<double>::CrossProduct_3(J,N,Iaux);
    Geom<double>::CrossProduct_3(I,J,N);
    Geom<double>::CrossProduct_3(J,N,I);
    Geom<double>::Normalize_3(N);
    Geom<double>::Normalize_3(I);
    Geom<double>::Normalize_3(J);

    double * O = mesh->Points()[i];
    
    std::vector<int>::const_iterator 
      itb = this->neighborhood->GetNeighborhoodPoints()[i].begin(), 
      ite = this->neighborhood->GetNeighborhoodPoints()[i].end();
    for(; itb != ite; itb++){
      double OP[3];
      double * P = mesh->Points()[(*itb)];

      Geom<double>::Substract_3(OP, P, O); 
      double h = Geom<double>::DotProduct_3(OP,N);
      double a = Geom<double>::DotProduct_3(OP,I);
      double b = Geom<double>::DotProduct_3(OP,J);

      double r = sqrt(a*a+b*b);
      double theta = atan2(b,a) + PI; // offset from -pi:pi to 0:2pi
      //cerr << r << " < " << this->neighborhood->GetDistance() << std::endl;
      //cerr << Geom<double>::VectorSize_3(OP) << " < " << this->neighborhood->GetDistance() << std::endl;


      int i_r = int (r/radiusStepSize);
      int i_theta = int(theta/angleStepSize);
      //cerr << i_r << " " << i_theta << std::endl;
      //cerr << radiusBinNb << " " << angleBinNb << std::endl;
      if (this->values == NULL){
	histogram[i_r][i_theta] += h;
      }
      else{
	histogram[i_r][i_theta] += (*values)[(*itb)];
      }
      histocount[i_r][i_theta] += 1;    
    }
    
    // compute mean
    for(int p=0; p<radiusBinNb; p++){
      for(int q=0; q<angleBinNb; q++){
	if (histocount[p][q] != 0){
	  histogram[p][q] /= histocount[p][q];
	}
      }
    }


    LSD[i] = std::numeric_limits<double>::max();   
    //int argmin = -1;
    for(int q=0; q<angleBinNb/2; q++){
      double s = histogram.AbsDiffAxialSymetricSum(q);
      if (s <LSD[i]){
	LSD[i] = s;
	//argmin = q;
      }
    }
    //cc LSD[i] = std::min(LSD[i],30.0);

    //double alpha = argmin*angleStepSize;
    //double A[3],B[3],D[3]; 
    //Geom<double>::Mult_3(A,sin(alpha),J);
    //Geom<double>::Mult_3(B,cos(alpha),I);
    //Geom<double>::Add_3(D,A,B);
    //
    //double X[] = {1.0,0.0,0.0};
    //double Y[] = {0.0,1.0,0.0};
    //double Z[] = {0.0,0.0,1.0};
    //LSD[i] = fabs(Geom<double>::DotProduct_3(Z,D));
    //
    //LSD[i] = argmin;




  }  
}

