//
// Copyright (C) 2010  Clement Creusot
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef GEOMH
#define GEOMH

#include <cstdio>

#include <math.h>
#include <vector>
#include <ostream>
#include <stdlib.h>
#include <utility> // comparison between pairs
#include <algorithm> // sorting

#define PI 3.1415926535897932384
#define EPSILON 0.000001

// calcul des normal
#define getk1t1(a00,a10) 6.0*(a00*(a00*a00*ab[4]+a10*a10*ab[6])+a10*(a00*a00*ab[5]+a10*a10*ab[7]));

// this are very low level function, that will be called in inner loops
// we write them inline, and dont use any object oriented indirection 

// Precision  float, double...

template < class Precision >
class Geom{

 public:


  typedef Precision Vector3[3];
  typedef Precision Vector4[4];


  class Line{
  public:
    Line(Precision* o, Precision* d){
      Copy_3(o,orig);
      Copy_3(d,dir);
    };
    

    Vector3 orig;
    Vector3 dir;
  };


  static inline void setInter(Vector3 out,Precision alpha,Precision beta,Vector3 in1,Vector3 in2){
    out[0] = (alpha*in2[0]+beta*in1[0])/(alpha+beta);
    out[1] = (alpha*in2[1]+beta*in1[1])/(alpha+beta);
    out[2] = (alpha*in2[2]+beta*in1[2])/(alpha+beta);
  }

  static inline void setCenter(Vector3 out,Vector3 in1,Vector3 in2,Vector3 in3){
    out[0] = (in1[0]+in2[0]+in3[0])/3.0;
    out[1] = (in1[1]+in2[1]+in3[1])/3.0;
    out[2] = (in1[2]+in2[2]+in3[2])/3.0;
  }


  static inline bool IsDirect_3(Vector3 x,Vector3 y,Vector3 z){
    Vector3 xy;
    CrossProduct_3(xy, x, y); 
    return (DotProduct_3(xy,z) >0);
  }
  
  static inline void Switch_3(Vector3 v1,Vector3 v2){
    Vector3 tmp;
    Copy_3(v1, tmp);
    Copy_3(v2, v1);
    Copy_3(tmp, v2);
  }

  static inline void Switch_1(Precision* v1,Precision* v2){
    Precision tmp;
    tmp = *v1;
    *v1 = *v2;
    *v2 = tmp;
  }

  static inline void Copy_3(const Vector3 in, Vector3 out){
    out[0] = in[0];
    out[1] = in[1];
    out[2] = in[2];
  }

  static inline void Copy_4(const Precision * in, Precision * out){
    out[0] = in[0];
    out[1] = in[1];
    out[2] = in[2];
    out[3] = in[3];
  }

  static inline Precision VectorSize_3(const Vector3 in){
    return sqrt(((in[0])*(in[0])+(in[1])*(in[1])+(in[2])*(in[2])));
  }
  
  static inline Precision Size_3(const Vector3 in){
    return sqrt(((in[0])*(in[0])+(in[1])*(in[1])+(in[2])*(in[2])));
  }

  static inline Precision Size_4(const Vector4 in){
    return sqrt(((in[0])*(in[0])+(in[1])*(in[1])+(in[2])*(in[2])+(in[3])*(in[3])));
  }


  static inline bool IsEqual_3(const Vector3 in1, const Vector3 in2){
    return in1[0]==in2[0] && in1[1]==in2[1] && in1[2]==in2[2];
  }

  // Compute the position of a point x between a and b
  // if ratio is between 0 and 1 and correspond to |ax|/|ab|
  static inline void MiddlePoint_3(Vector3 out, const Vector3 a, const Precision ratio, const Vector3 b){
    out[0] = (1-ratio)*a[0] + ratio*b[0];
    out[1] = (1-ratio)*a[1] + ratio*b[1];
    out[2] = (1-ratio)*a[2] + ratio*b[2];
  }

  static inline void Divide_3(Vector3 inout, Precision value){
    Precision v = (value ==0)? 1.0: value; 
    inout[0] = inout[0]/v;
    inout[1] = inout[1]/v;
    inout[2] = inout[2]/v;
  }

  static inline void Divide_4(Vector4 inout, Precision value){
    Precision v = (value ==0)? 1.0: value; 
    inout[0] = inout[0]/v;
    inout[1] = inout[1]/v;
    inout[2] = inout[2]/v;
    inout[3] = inout[3]/v;
  }


  static inline void Substract_3(Vector3 out, const Vector3 in1, const Vector3 in2){
    out[0] = in1[0]-in2[0];
    out[1] = in1[1]-in2[1];
    out[2] = in1[2]-in2[2];
  }

  static inline void Add_3(Vector3 out, const Vector3 in1, const Vector3 in2){
    out[0] = in1[0]+in2[0];
    out[1] = in1[1]+in2[1];
    out[2] = in1[2]+in2[2];
  }
  static inline void Add_3(Vector3 inout, const Vector3 in1){
    inout[0] += in1[0];
    inout[1] += in1[1];
    inout[2] += in1[2];
  }

  static inline void Add_4(Vector4 inout, const Vector4 in1){
    inout[0] += in1[0];
    inout[1] += in1[1];
    inout[2] += in1[2];
    inout[3] += in1[3];
  }


  static inline void Mult_3(Vector3 out, const Precision scal, const Vector3 in){
    out[0] = scal*in[0];
    out[1] = scal*in[1];
    out[2] = scal*in[2];
  }

  
  static inline void Middle_3(Vector3 out, const Vector3 in1, const Vector3 in2){
    out[0] = (in1[0]+in2[0])/2.0;
    out[1] = (in1[1]+in2[1])/2.0;
    out[2] = (in1[2]+in2[2])/2.0;
  }

  static inline Precision Distance_3(const Vector3 in1, const Vector3 in2){
    Precision tmp[3];
    Substract_3(tmp, in1, in2);
    return VectorSize_3(tmp);
  }

  static inline Precision TriangleSurface_3(const Vector3 in1, const Vector3 in2, const Vector3 in3){
    Precision xy1[3] = {in1[0], in1[1], 1};
    Precision xy2[3] = {in2[0], in2[1], 1};
    Precision xy3[3] = {in3[0], in3[1], 1};
    Precision yz1[3] = {in1[1], in1[2], 1};
    Precision yz2[3] = {in2[1], in2[2], 1};
    Precision yz3[3] = {in3[1], in3[2], 1};
    Precision zx1[3] = {in1[2], in1[0], 1};
    Precision zx2[3] = {in2[2], in2[0], 1};
    Precision zx3[3] = {in3[2], in3[0], 1};
    Precision detz = Det_3(xy1,xy2,xy3);
    Precision detx = Det_3(yz1,yz2,yz3);
    Precision dety = Det_3(zx1,zx2,zx3);
    return 0.5 * sqrt(detz*detz+detx*detx+dety*dety);
  }

  static inline Precision TetraVolume_3(const Vector3 in1, const Vector3 in2, const Vector3 in3, const Vector3 in4){
    Precision v1[3], v2[3], v3[3];
    Substract_3(v1, in1, in2);
    Substract_3(v2, in2, in3);
    Substract_3(v3, in3, in4);
    return fabs(Det_3(v1,v2,v3))/6;
  }



  static inline Precision Det_3(const Vector3 in1, const Vector3 in2, const Vector3 in3){
    return (in1[0]*in2[1]*in3[2] - in1[0]*in2[2]*in3[1] +
	    in1[1]*in2[2]*in3[0] - in1[1]*in2[0]*in3[2] +
	    in1[2]*in2[0]*in3[1] - in1[2]*in2[1]*in3[0]);
  }
  

  static inline void CrossProduct_3(Vector3 out, const Vector3 in1, const Vector3 in2){
    out[0] = ((in1[1])*(in2[2]) - (in2[1])*(in1[2]));
    out[1] = ((in1[2])*(in2[0]) - (in2[2])*(in1[0]));
    out[2] = ((in1[0])*(in2[1]) - (in2[0])*(in1[1]));
  }
  

  static inline Precision RealSolidAngle_3(const Vector3 a, const Vector3 b, const Vector3 c){  
    Precision det = fabs(Det_3(a,b,c));
    
    Precision al = VectorSize_3(a);
    Precision bl = VectorSize_3(b);
    Precision cl = VectorSize_3(c);
    
    Precision div = al*bl*cl + Dot_3(a,b)*cl + Dot_3(a,c)*bl + Dot_3(b,c)*al;
    Precision at = atan2(det, div);
    if(at < 0) at += PI; // If det>0 && div<0 atan2 returns < 0, so add pi.
    Precision omega = 2.0 * at;
    
    return omega;
  }

  // assume normalized vectors
  static inline Precision SolidAngle_3(const Vector3 in1, const Vector3 in2, const Vector3 in3){  
    return fabs(Det_3(in1,in2,in3));
  }

  static inline Precision SolidAngle_3(const Vector3 O, const Vector3 A, const Vector3 B, const Vector3 C){  
    Precision OA[3],OB[3],OC[3];
    Substract_3(OA,A,O);
    Substract_3(OB,B,O);
    Substract_3(OC,C,O);
    Normalize_3(OA);
    Normalize_3(OB);
    Normalize_3(OC);
    return SolidAngle_3(OA,OB,OC);
  }


  
  static inline void TriangleNormal_3(Vector3 out, const Vector3 in1, const Vector3 in2,  const Vector3 in3){
    Precision v1[3], v2[3];
    Substract_3(v1, in2, in1);
    Substract_3(v2, in3, in1);
    CrossProduct_3(out, v1,v2);
    Normalize_3(out);
  }

  // plane define by three points
  static inline Precision DistanceToPlan_3(const Vector3 pts, const Vector3 in1, const Vector3 in2,  const Vector3 in3){
    Precision v[3], n[3];
    Substract_3(v, pts, in1);
    TriangleNormal_3(n, in1, in2, in3);
    return DotProduct_3(n,v);
  }
  
  // plane define by one point and one normal (already normalized)
  static inline Precision DistanceToPlan_3(const Vector3 pts, const Vector3 O, const Vector3 N){
    Precision v[3];
    Substract_3(v, pts, O);
    return DotProduct_3(N,v);
  }

  
  
  static inline Precision DotProduct_3(const Vector3 in1, const Vector3 in2){
    return ((in1[0])*(in2[0]) + (in1[1])*(in2[1]) + (in1[2])*(in2[2]));
  }

  static inline Precision DotProduct_4(const Vector4 in1, const Vector4 in2){
    return ((in1[0])*(in2[0]) + (in1[1])*(in2[1]) + (in1[2])*(in2[2]) + (in1[3])*(in2[3]));
  }

  
  static inline void setVirtualTangents(const Vector3 normal, Vector3 outt1,Vector3 outt2){
    // create a orthonormalized base from the vector Normal.
    
    if(fabs(normal[0])<fabs(normal[1])){
      outt1[0] = 0.0;
      outt1[1] = -normal[2];
      outt1[2] = normal[1];
    
    }else{
      outt1[0] = normal[2];
      outt1[1] = 0.0;
      outt1[2] = -normal[0];
    }
    Normalize_3(outt1);
    CrossProduct_3(outt2,normal,outt1);
    Normalize_3(outt2);
  }

  // WARNING: you changing the vector in place, be sure its a copy
  // if you expect to use the vector afterwards.
  static inline void Normalize_3(Vector3 inout){
    Precision dsize = VectorSize_3(inout);
    if(dsize==0.0)dsize=1.0;
    dsize = 1/dsize;
    inout[0] *= dsize;
    inout[1] *= dsize;
    inout[2] *= dsize;
  }


  static inline void Normalize_4(Vector4 inout){
    Precision dsize = Size_4(inout);
    if(dsize>0.00000001){
      dsize = 1/dsize;
      inout[0] *= dsize;
      inout[1] *= dsize;
      inout[2] *= dsize;
      inout[3] *= dsize;
    }
  }

  
  // FIXME do unit test for the geometric functions
  // interpolate the value at point p wich belong to the triangle
  // k1,k2,k3.
  static inline Precision Interpolation_3(const Vector3 p,  
					  const Vector3 k1, const Precision v1,
					  const Vector3 k2, const Precision v2, 
					  const Vector3 k3, const Precision v3){
    Precision Xa[3], Xb[3], V[3];
    Substract_3(Xa, k2, k1);
    Substract_3(Xb, k3, k1);
    Substract_3(V, p, k1);
    
    Precision normXa = VectorSize_3(Xa);
    Precision normXb = VectorSize_3(Xb);

    Precision a = DotProduct_3(Xa, V)/(normXa * normXa);
    Precision b = DotProduct_3(Xb, V)/(normXb * normXb);

    return (v1 + (v2-v1)*a + (v3-v1)*b);

  }


  
  static void LinearRegression( Vector3 P, Vector3 V, const std::vector<Precision*>& points, int start = 0, int end = -1){
    Precision sumX, sumY, sumZ;
    Precision sumZZ, sumZY, sumZX; 
    sumX = sumY  = sumZ = 0.0;
    sumZZ = sumZY = sumZX = 0.0;
    int pointNb = (int)points.size();
    
    if (! points.empty()){
     
      // 'option' allow to select wich part of the set is selected
      // for the regression. 
      // positif number means the N first points
      // negatif number means the N last points
      // 0 means all the points
      int i = start;
      int n = end;
      if (end == -1 or end > pointNb) {
	n = pointNb;
      }
      if (start <0){
	i = 0;
      }
      if (start > pointNb){
	i = n;
      }


      for (; i!=n ; i++){
	Precision x = points[i][0];
	Precision y = points[i][1];
	Precision z = points[i][2];
	sumX += x;
	sumY += y;
	sumZ += z;
	sumZZ += z * z;
	sumZX += x * z;
	sumZY += y * z;
      }
     

      if (n >= 2){
	// grosso modo on cherche P(X0,Y0,Z0) et V(Vx,Vy,Vz) tel que
	// la courbe vale P + V*t
	// On a trop dinconue on pose donc Z0 = 0, VZ = 1
	// On calcul la distance a la courbe de chaque point
	// On pose que la somme doit etre minimal
	// On prend donc la derive egal a zero
	// On develope et on a 4 equations

	// FIXME: if sumZZ = 0 then our assumptions are wrong
	// because the line will never cross the plan Z=0
	// Use a similar computation with sumYY or sumXX
	// in an if/elseif/else
	if (sumZZ == 0.0){
	  fprintf(stderr, "Not yet implemented case\n");
	  exit(1);
	}
	
	// Compute two factors to avoid to much calculus
	Precision factor1 = sumZ /(sumZZ * n);
	Precision factor2 = 1/(1- sumZ*factor1); 

	// Determine the direction Vector
	V[0] = (sumZX/sumZZ - sumX*factor1)*factor2; //Vx
	V[1] = (sumZY/sumZZ - sumY*factor1)*factor2; //Vy
	V[2] = 1; //Vz
	
	// Dertermine a starting point along the line
	P[0] = (sumX/n - sumZX*factor1)*factor2; //X0
	P[1] = (sumY/n - sumZY*factor1)*factor2; //Y0
	P[2] = 0; //Z0
	
	// normalize the direction vector
	Normalize_3(V);
      }
    }
  }  
  
  // compute the distance between the point P1 and the line define by
  // PO and the unit direction V0
  static inline Precision DistanceToLine_3(Vector3 P1, Vector3 P0, Vector3 V0){
    Precision P0P1[3];
    Precision T[3], N[3];
    Substract_3(P0P1, P1, P0);
    CrossProduct_3(N, V0, P0P1);     
    CrossProduct_3(T, N, V0);
    return VectorSize_3(T);      
  }


  // compute the two points where the distance between two line become minimal
  // O1 = P1 + alpha V1
  // O2 = P2 + beta V2
  // System:
  // | (O1-O2).V1 = 0
  // | (O1-O2).V2 = 0
  // develope:
  // | (P1-P2).V1 + alpha - beta V1.V2 = 0
  // | (P1-P2).V2 + alpja V1.V2 - beta = 0
  // dou:
  // | alpha = ((P1-P2).(V1.V2 * V2 - V1))/(1-(V1.V2)^2)
  // | beta = (P1-P2)V2 + alpha V1.V2
  static inline Precision SmallestSegmentBetweenLine_3(Vector3 P1, Vector3 V1, Vector3 P2, Vector3 V2, Precision* O1, Precision* O2){

   Precision P1P2[3];
   Precision numer,denom;
   Substract_3(P1P2, P1, P2);// P1P2 eq (P1-P2) not P1->P2  
   Precision dP1P2_V2 = DotProduct_3(P1P2, V2);     
   Precision dV2_V1 = DotProduct_3(V2, V1);     
   Precision dP1P2_V1 = DotProduct_3(P1P2, V1);     

   denom = 1  - dV2_V1 * dV2_V1;
   // test colinearity
   if (fabs(denom) < 0.001){
     return false;
   }
   numer = dP1P2_V2 * dV2_V1 - dP1P2_V1 ;

   Precision alpha = numer / denom;
   Precision beta = (dP1P2_V2 + dV2_V1 * alpha);

   Mult_3(O1,alpha,V1);
   Add_3(O1,P1);
   Mult_3(O2,beta,V2);
   Add_3(O2,P2);
   
   return true;
  }



  // compute the projection of P1 on the line define by
  // PO and the unit direction V0
  static inline void ProjectionOnLine_3(Vector3 out, Vector3 P1, Vector3 P0, Vector3 V0 ){
    Precision P0P1[3];
    Precision T[3], N[3];
    Substract_3(P0P1, P1, P0);
    CrossProduct_3(N, V0, P0P1);     
    CrossProduct_3(T, N, V0);
    Substract_3(out, P1, T);      
  }


  // define the plan that contains v1/\v2 and wich projection along v1/\v2 is the bisection of the line projection of v1 and the line projection of v2
  // FIXME the calcul for the point belonging to the plan are just the brut result copied from maxima solution of :
  // (X1-P1)/\V1 = 0, (X2-P2)/\V2 = 0, (X1-X2)/\(V1/\V2) =0
  // where X1 and X2 are the unknown points on D1 and D2
  // Here we only compute X1 (Pout).
  // some improvement are needed.
  // the output is a point belonging to the plan plus a normal to this plan
  static inline void BisectionPlan(Vector3 Pout   , Vector3 Nout, 
				   const Vector3 p1, const Vector3 v1, 
				   const Vector3 p2, const Vector3 v2){

    // Compute one point on the bisection plan
    Precision diviseur = ((v1[1]*v1[1]+v1[0]*v1[0])*v2[2]*v2[2]+(-2*v1[1]*v1[2]*v2[1]-2*v1[0]*v1[2]*v2[0])*v2[2]+(v1[2]*v1[2]+v1[0]*v1[0])*v2[1]*v2[1]-2*v1[0]*v1[1]*v2[0]*v2[1]+(v1[2]*v1[2]+v1[1]*v1[1])*v2[0]*v2[0]);
  
    Pout[0] = ((p1[0]*v1[1]*v1[1]+(p2[1]-p1[1])*v1[0]*v1[1]+p2[0]*v1[0]*v1[0])*v2[2]*v2[2]+((((p1[1]-p2[1])*v1[0]-2*p1[0]*v1[1])*v1[2]+(p1[2]-p2[2])*v1[0]*v1[1])*v2[1]+((-p2[0]-p1[0])*v1[0]*v1[2]+(p1[2]-p2[2])*v1[0]*v1[0])*v2[0])*v2[2]+(p1[0]*v1[2]*v1[2]+(p2[2]-p1[2])*v1[0]*v1[2]+p2[0]*v1[0]*v1[0])*v2[1]*v2[1]+((-p2[0]-p1[0])*v1[0]*v1[1]+(p1[1]-p2[1])*v1[0]*v1[0])*v2[0]*v2[1]+(p1[0]*v1[2]*v1[2]+(p2[2]-p1[2])*v1[0]*v1[2]+p1[0]*v1[1]*v1[1]+(p2[1]-p1[1])*v1[0]*v1[1])*v2[0]*v2[0]);
  
    Pout[1] = ((p2[1]*v1[1]*v1[1]+(p2[0]-p1[0])*v1[0]*v1[1]+p1[1]*v1[0]*v1[0])*v2[2]*v2[2]+(((-p2[1]-p1[1])*v1[1]*v1[2]+(p1[2]-p2[2])*v1[1]*v1[1])*v2[1]+(((p1[0]-p2[0])*v1[1]-2*p1[1]*v1[0])*v1[2]+(p1[2]-p2[2])*v1[0]*v1[1])*v2[0])*v2[2]+(p1[1]*v1[2]*v1[2]+(p2[2]-p1[2])*v1[1]*v1[2]+(p2[0]-p1[0])*v1[0]*v1[1]+p1[1]*v1[0]*v1[0])*v2[1]*v2[1]+((p1[0]-p2[0])*v1[1]*v1[1]+(-p2[1]-p1[1])*v1[0]*v1[1])*v2[0]*v2[1]+(p1[1]*v1[2]*v1[2]+(p2[2]-p1[2])*v1[1]*v1[2]+p2[1]*v1[1]*v1[1])*v2[0]*v2[0]);
  
    Pout[2] = ((((p2[1]-p1[1])*v1[1]+(p2[0]-p1[0])*v1[0])*v1[2]+p1[2]*v1[1]*v1[1]+p1[2]*v1[0]*v1[0])*v2[2]*v2[2]+(((p1[1]-p2[1])*v1[2]*v1[2]+(-p2[2]-p1[2])*v1[1]*v1[2])*v2[1]+((p1[0]-p2[0])*v1[2]*v1[2]+(-p2[2]-p1[2])*v1[0]*v1[2])*v2[0])*v2[2]+(p2[2]*v1[2]*v1[2]+(p2[0]-p1[0])*v1[0]*v1[2]+p1[2]*v1[0]*v1[0])*v2[1]*v2[1]+(((p1[0]-p2[0])*v1[1]+(p1[1]-p2[1])*v1[0])*v1[2]-2*p1[2]*v1[0]*v1[1])*v2[0]*v2[1]+(p2[2]*v1[2]*v1[2]+(p2[1]-p1[1])*v1[1]*v1[2]+p1[2]*v1[1]*v1[1])*v2[0]*v2[0]);

  
    Divide_3(Pout, diviseur);

    // Compute the normal to the plan
    Precision T1[3], T2[3];
    CrossProduct_3(T1, v1, v2);
    Add_3(T2, v1, v2);
    CrossProduct_3(Nout, T1, T2);
    Normalize_3(Nout);
  }

  //test few simple cases
  bool testBisectionPlan(){
    bool pass = true;
    Precision V1[3], V2[3];
    Precision P1[3], P2[3];
    Precision Pout[3], Nout[3];
  
    P1[0] = 2; P1[1] = 0; P1[2] = 0;
    V1[0] = 1; V1[1] = 0; V1[2] = 0;
    P2[0] = 1; P2[1] = 1; P2[2] = 1;
    V2[0] = 0; V2[1] = 0; V2[2] = 1;
    BisectionPlan(Pout,Nout,P1,V1,P2,V2);
    if( !(Pout[0] == 1 && Pout[1] == 0 && Pout[2] == 0) )
      pass = false; 
    if( !(Nout[0] == -Nout[2] && Nout[1] == 0) )
      pass = false;

    return pass;
  }

  


  // 4X4 functions
  static inline void Multiply4x4_point_3(Precision* mat, Precision* pts, Precision* out){
    out[0] = pts[0]*mat[0] +  pts[1]*mat[1] +  pts[2]*mat[2] +  1*mat[3]; 
    out[1] = pts[0]*mat[4] +  pts[1]*mat[5] +  pts[2]*mat[6] +  1*mat[7]; 
    out[2] = pts[0]*mat[8] +  pts[1]*mat[9] +  pts[2]*mat[10]+  1*mat[11]; 
  }
  static inline void Multiply4x4_vector_3(Precision* mat, Precision* pts, Precision* out){
    out[0] = pts[0]*mat[0] +  pts[1]*mat[1] +  pts[2]*mat[2] ;
    out[1] = pts[0]*mat[4] +  pts[1]*mat[5] +  pts[2]*mat[6] ;
    out[2] = pts[0]*mat[8] +  pts[1]*mat[9] +  pts[2]*mat[10]; 
  }


  
  static inline void minmax_3(Precision a1, Precision a2, Precision a3, 
			      Precision& min, Precision& max){
    if (a1 > a2){
      if( a2 > a3){
	max = a1; min = a3;
      }
      else if(a1 > a3){
	max = a1; min = a2;
      } 
      else{
	max = a3; min = a2;
      }
    }
    else{
      if( a1 > a3){
	max = a2; min = a3;
      }
      else if(a2 > a3){
	max = a2; min = a1;
      } 
      else{
	max = a3; min = a1;
      }
    }
  } 

  static inline Precision MinTetraEdgeSize_3(Vector3 a1, Vector3 a2, 
					Vector3 a3, Vector3 a4){
    Precision value = Geom<Precision>::Distance_3(a1,a2);
    value = std::min(value, Geom<Precision>::Distance_3(a1,a3));
    value = std::min(value, Geom<Precision>::Distance_3(a1,a4));
    value = std::min(value, Geom<Precision>::Distance_3(a2,a3));
    value = std::min(value, Geom<Precision>::Distance_3(a2,a4));
    value = std::min(value, Geom<Precision>::Distance_3(a3,a4));
    return value;
  }

  static inline Precision MaxTetraEdgeSize_3(Vector3 a1, Vector3 a2, 
					     Vector3 a3, Vector3 a4){
    Precision value = Geom<Precision>::Distance_3(a1,a2);
    value = std::max(value, Geom<Precision>::Distance_3(a1,a3));
    value = std::max(value, Geom<Precision>::Distance_3(a1,a4));
    value = std::max(value, Geom<Precision>::Distance_3(a2,a3));
    value = std::max(value, Geom<Precision>::Distance_3(a2,a4));
    value = std::max(value, Geom<Precision>::Distance_3(a3,a4));
    return value;
  }

  static inline Precision MinTriangleEdgeSize_3(Vector3 a1, Vector3 a2, 
						Vector3 a3){
    Precision value = Geom<Precision>::Distance_3(a1,a2);
    value = std::min(value, Geom<Precision>::Distance_3(a1,a3));
    value = std::min(value, Geom<Precision>::Distance_3(a2,a3));
    return value;
  }

  static inline Precision MaxTriangleEdgeSize_3(Vector3 a1, Vector3 a2, 
						Vector3 a3){
    Precision value = Geom<Precision>::Distance_3(a1,a2);
    value = std::max(value, Geom<Precision>::Distance_3(a1,a3));
    value = std::max(value, Geom<Precision>::Distance_3(a2,a3));
    return value;
  }

  
  static inline 
    void TriangleAngles_3(Vector3 a1, Vector3 a2, Vector3 a3,
			  Precision& t1, Precision& t2, Precision& t3){
    Precision v1[3], v2[3], v3[3];
    Substract_3(v1, a2, a1);
    Substract_3(v2, a3, a1);
    Substract_3(v3, a3, a2);
    Normalize_3(v1);
    Normalize_3(v2);
    Normalize_3(v3);
    t1 = acos(DotProduct_3(v1,v2));    
    t2 = acos(-DotProduct_3(v1,v3));    
    t3 = acos(DotProduct_3(v2,v3));    
  }

  static inline 
    Precision MinTriangleAngles_3(Vector3 a1, Vector3 a2, Vector3 a3){
    Precision t1,t2,t3;
    TriangleAngles_3(a1, a2, a3, t1, t2, t3);
    return min(min(t1,t2),t3);
  }
  

  // Get the x,y,z bounding box of a triangle
  static inline void GetBoundingBox_3(Vector3 a, Vector3 b, Vector3 c, 
				      Precision& xmin, Precision& xmax,
				      Precision& ymin, Precision& ymax,
				      Precision& zmin, Precision& zmax){
    minmax_3(a[0],b[0],c[0],xmin,xmax);
    minmax_3(a[1],b[1],c[1],ymin,ymax);
    minmax_3(a[2],b[2],c[2],zmin,zmax);
  }



   static inline bool IsLineIntersectingTriangle(Vector3 orig, Vector3 dir, Vector3 a , Vector3 b, Vector3 c , Vector3 inter ){
    Precision  u[3], v[3], n[3];             
    Precision  w0[3], w[3];          
    Precision  r, scal1, scal2;             
    
    Substract_3(u,b,a);
    Substract_3(v,c,a);
    CrossProduct_3(n,u,v); 
        
    Substract_3(w0,orig,a);
    scal1 = -DotProduct_3(n,w0);
    scal2 = DotProduct_3(n,dir);
   
    r = scal1 / scal2;

    Mult_3(dir, r, dir);
    Add_3(inter,orig,dir);           
    
    Precision    uu, uv, vv, wu, wv, D;
    uu = DotProduct_3(u,u);
    uv = DotProduct_3(u,v);
    vv = DotProduct_3(v,v); 
    Substract_3(w,inter,a);   
    wu = DotProduct_3(w,u);
    wv = DotProduct_3(w,v);
    D = uv * uv - uu * vv;

    Precision s, t;
    s = (uv * wv - vv * wu) / D;
    if (s < 0.0 || s > 1.0)        
      return false;
    t = (uv * wu - uu * wv) / D;
    if (t < 0.0 || (s + t) > 1.0)  
      return false;
    
    return true;                      
  }


   // see wikipedia article on trilateration
   // return number of intersection 0 1 or 2
   // put result in res1 res2
   static inline int Trilateration(Vector3 p1, Precision r1,
				   Vector3 p2, Precision r2,
				   Vector3 p3, Precision r3,
				   Vector3 res1, Vector3 res2){
     Vector3 v1,v2,v3;
     Substract_3(v1, p2, p1);     
     Substract_3(v2, p3, p1);    
     Substract_3(v3, p2, p3);    
     Precision d1,d2,d3;
     d1 = VectorSize_3(v1);
     d2 = VectorSize_3(v2);
     d3 = VectorSize_3(v3);

     Precision i = (d2*d2-d3*d3+d1*d1)/(2*d1);
     if (d2*d2-i < 0){
       return 0;
     }
     Precision j = - sqrt(d2*d2-i*i);// imply CAB angle positive trigo

     Precision x = (r1*r1 - r2*r2 + d1*d1)/(2*d1);
     Precision y = (r1*r1 - r3*r3 + i*i + j*j)/(2*j) - i/j*x;
     
     Precision aux = r1*r1 - x*x - y*y; 

     if (aux < 0){
       return 0;
     }

     Precision z1 = + sqrt(aux);
     Precision z2 = - sqrt(aux);

     // compute image of x,y,z in the original space
     Normalize_3(v1); // Ximage
     Vector3 Zimage;
     CrossProduct_3(Zimage, v2, v1);
     Normalize_3(Zimage);
     Vector3 Yimage;
     CrossProduct_3(Yimage, Zimage, v1);
 
     Vector3 AuxVect;
     Mult_3(v1,x,v1);
     Mult_3(Yimage,y,Yimage);
     Add_3(AuxVect,v1,Yimage);
     Add_3(AuxVect,p1);
     
     Mult_3(res1,z1,Zimage);
     Mult_3(res2,z2,Zimage);
     Add_3(res1,AuxVect);
     Add_3(res2,AuxVect);
     
     return 1 + (int) (aux!=0);
   }
   // test function
   // #include <cstdlib> // for rand
   // for (int i=0;i<100;i++){
   //    Precision  A[] = {(Precision)(rand()%100),(Precision)(rand()%100),(Precision)(rand()%100)};
   //    Precision  B[] = {(Precision)(rand()%100),(Precision)(rand()%100),(Precision)(rand()%100)};
   //    Precision  C[] = {(Precision)(rand()%100),(Precision)(rand()%100),(Precision)(rand()%100)};
   //    Precision  X[] = {(Precision)(rand()%100),(Precision)(rand()%100),(Precision)(rand()%100)}; 
   //    Precision dAX = Geom<Precision>::Distance_3(A,X);
   //    Precision dBX = Geom<Precision>::Distance_3(B,X);
   //    Precision dCX = Geom<Precision>::Distance_3(C,X);
   //    Precision res1[3],res2[3];
   //    Geom<Precision>::Trilateration(A,dAX,B,dBX,C,dCX,res1,res2);
   //    if (Geom<Precision>::Distance_3(res1,X) >0.1 && Geom<Precision>::Distance_3(res2,X) >0.1){
   //      fprintf(stderr,"PROBLEM %d\n",i);
   //      fprintf(stderr, "res1: %f %f %f\n",res1[0],res1[1],res1[2]);
   //      fprintf(stderr, "res2: %f %f %f\n",res2[0],res2[1],res2[2]);
   //      fprintf(stderr, "X: %f %f %f\n",X[0],X[1],X[2]);
   //      
   //    }
   //  }


   static inline Precision DistanceToPlane_3(Vector3 in, 
					     Vector3 O, Vector3 N){
     // P(x,y,z): projection of IN in the plane
     // P+ t*N = IN
     // a(x-x0) + b(y-y0) + c(z-z0) = 0
     
     Precision V[3];
     Substract_3(V,in,O);
     return DotProduct_3(N,V)/DotProduct_3(N,N);
   }
   
   static inline void SymetricPointToPlane_3(Vector3 out, Vector3 in, 
					     Vector3 O, Vector3 N){
     Precision V[3];
     Precision t = DistanceToPlane_3(in,O,N);
     //OUT = IN - 2*t*N
     Mult_3(V,2*t,N);
     Substract_3(out,in,V);
   }

   static inline void ProjectPointInPlane_3(Vector3 out, Vector3 in, 
					    Vector3 O, Vector3 N){
     Precision V[3];
     Precision t = DistanceToPlane_3(in,O,N);
     //OUT = IN - t*N
     Mult_3(V,t,N);
     Substract_3(out,in,V);
   }

   // project on plane(O,N) along direction U
   static inline void ProjectPointInPlaneAlongLine_3(Vector3 out, Vector3 in, 
						     Vector3 O, Vector3 N,
						     Vector3 U){
     // OUT = (N/\(IN/\U) -d*U) / (N.U)
     Precision V[3];
     CrossProduct_3(V,in,U);
     CrossProduct_3(out,N,V);
     Mult_3(V,DotProduct_3(N,O), U);//-d*U
     Add_3(out, V);
     Mult_3(out, 1/DotProduct_3(N,U), out);
   }



   // compute alpha,beta,gamma such that X = alpha*OA + beta*OB + gamma*OC
   // we assume no coordinate is known, only the relative distances
   static inline bool RelativeVector(Precision dOA, Precision dOB, Precision dOC, Precision dOX,  
				     Precision dAB, Precision dAC, Precision dAX, 
				     Precision dBC, Precision dBX, Precision dCX,
				     Precision& alpha, Precision& beta, Precision& gamma){ 
     
     Vector3 O = {0.0, 0.0, 0.0};
     Vector3 X = {0.0, 0.0, 0.0};
     Vector3 A = {dOA , 0.0, 0.0};
     Precision i = (dOB*dOB - dAB*dAB + dOA*dOA)/(2*dOA);
     Precision j = sqrt(dOB*dOB - i*i);
     Vector3 B = {i  , j, 0.0};

     Precision res1[3], res2[3];
     int ret1 = Trilateration(O, dOC, A, dAC, B, dBC, res1, res2);
     Precision res3[3], res4[3];
     int ret2 = Trilateration(O, dOX, A, dAX, B, dBX, res3, res4);
     if (ret1==0 || ret2 == 0) return false;
     Precision C[3];
     Copy_3(res1,C);

     Precision min = 1.0;
     if (abs(dCX-Distance_3(C,res3)) <min){
       Copy_3(res3,X); 
       min = abs(dCX-Distance_3(C,res3));
     }
     if (abs(dCX-Distance_3(C,res4)) <min){
       Copy_3(res4,X);
       min = abs(dCX-Distance_3(C,res4));
     }
     if (min == 0.1){
       fprintf(stderr, "Problem CX:%f ret1=%d ret2=%d\n",dCX,ret1,ret2);
       fprintf(stderr, "res1: %f %f %f\n",res1[0],res1[1],res1[2]);
       fprintf(stderr, "res2: %f %f %f\n",res2[0],res2[1],res2[2]);
       fprintf(stderr, "res3: %f %f %f\n",res3[0],res3[1],res3[2]);
       fprintf(stderr, "res4: %f %f %f\n",res4[0],res4[1],res4[2]);
       fprintf(stderr, "1-3:%f, 1-4:%f, 2-3:%f, 2-4:%f\n",
	       Geom<Precision>::Distance_3(res1,res3),
	       Geom<Precision>::Distance_3(res1,res4),
	       Geom<Precision>::Distance_3(res2,res3),
	       Geom<Precision>::Distance_3(res2,res4));
       return false;
     }

     Precision Xp[3]; //Xp = X _|_ OAB along OC
     Vector3 Noab = {0.0, 0.0, 1.0};
     ProjectPointInPlaneAlongLine_3(Xp, X, O, Noab, C);
     
     gamma = Distance_3(Xp,X)/VectorSize_3(C);

     beta = Xp[1]/B[1];
     alpha = (Xp[0] - beta*B[0])/A[0];
     
     ////test
     //Precision OA[3],OB[3],OC[3],OUT[3];
     //Substract_3(OA,A,O);
     //Substract_3(OB,B,O);
     //Substract_3(OC,C,O);
     //Mult_3(OA,alpha,OA);
     //Mult_3(OB,beta ,OB);
     //Mult_3(OC,gamma,OC);
     //Add_3(OUT,O,OA);
     //Add_3(OUT,OB);
     //Add_3(OUT,OC);
     //
     //if(Distance_3(OUT,X) >0.1){
     //  fprintf(stderr, "Xp: %f %f %f\n",Xp[0],Xp[1],Xp[2]);
     //  fprintf(stderr, "res: %f %f %f\n",OUT[0],OUT[1],OUT[2]);
     //  fprintf(stderr, "  X: %f %f %f\n",X[0],X[1],X[2]);
     //  fprintf(stderr, "NOT WORKING\n");
     //  exit(1);
     //}

     return true;
   }





   // compute alpha,beta,gamma such that X = alpha*OA + beta*OB + gamma*OC
   static inline bool RelativePosition(Vector3 O,Vector3 A,Vector3 B,Vector3 C,Vector3 X,
				       Precision& alpha, Precision& beta, Precision& gamma){ 
     
     Precision OA[3],OB[3],OC[3],OX[3];
     Substract_3(OA,A,O);
     Substract_3(OB,B,O);
     Substract_3(OC,C,O);
     Substract_3(OX,X,O);
     Precision Xp[3]; //Xp : projected point on the three plane 
     Precision Noab[3], Noac[3], Nobc[3];
     CrossProduct_3(Noab,OA,OB);
     Normalize_3(Noab);
     CrossProduct_3(Noac,OC,OA);
     Normalize_3(Noac);
     CrossProduct_3(Nobc,OB,OC);
     Normalize_3(Nobc);


     //cc not very efficient, can be improved
     // use change of coordinate like in RelativeVector()
     Precision Xpc[3],Xpb[3],Xpa[3];
     ProjectPointInPlaneAlongLine_3(Xp, X, O, Noab, OC);
     Substract_3(Xpc,X,Xp);
     gamma=(1-2*(int)(DotProduct_3(Xpc,OC)<0))*Distance_3(Xp,X)/VectorSize_3(OC);
     ProjectPointInPlaneAlongLine_3(Xp, X, O, Noac, OB);
     Substract_3(Xpb,X,Xp);
     beta=(1-2*(int)(DotProduct_3(Xpb,OB)<0))*Distance_3(Xp,X)/VectorSize_3(OB);
     ProjectPointInPlaneAlongLine_3(Xp, X, O, Nobc, OA);
     Substract_3(Xpa,X,Xp);
     alpha=(1-2*(int)(DotProduct_3(Xpa,OA)<0))*Distance_3(Xp,X)/VectorSize_3(OA);
     

     //////test
     //Precision OUT[3],V[3];
     //Mult_3(V,alpha,OA);
     //Add_3(OUT,O,V);
     //Mult_3(V,beta ,OB);
     //Add_3(OUT,V);
     //Mult_3(V,gamma,OC);
     //Add_3(OUT,V);
     //
     //if(Distance_3(OUT,X) >0.1){
     //  fprintf(stderr, "abc: %f %f %f\n",alpha,beta,gamma);
     //  fprintf(stderr, "OA:  %f %f %f\n",OA[0],OA[1],OA[2]);
     //  fprintf(stderr, "OB:  %f %f %f\n",OB[0],OB[1],OB[2]);
     //  fprintf(stderr, "OC:  %f %f %f\n",OC[0],OC[1],OC[2]);
     //  fprintf(stderr, "OX:  %f %f %f\n",OX[0],OX[1],OX[2]);
     //  fprintf(stderr, "res: %f %f %f\n",OUT[0],OUT[1],OUT[2]);
     //  fprintf(stderr, "  X: %f %f %f\n",X[0],X[1],X[2]);
     //  fprintf(stderr, "NOT WORKING\n");
     //  exit(1);
     //}

     return true;     
   }


   static inline std::ostream& PrintMatrix4x4(std::ostream& stream,Precision mat[16]){
     for (int i=0; i<4; i++){
       if (i != 0){
	 stream << ";" ;
       }
       for (int j=0; j<4; j++){
	 if (j != 0){
	   stream << "," ;
	 }
	 stream << mat[i*4+j] ;
       }
     }
     return stream;
   }

   
   static inline void Matrix4x4_2_Matrix3x3(Precision * matrix4, Precision * matrix3){
     Copy_3(matrix4,matrix3);
     Copy_3(matrix4+4,matrix3+3);
     Copy_3(matrix4+8,matrix3+6);
   }

   static void Matrix4x4_2_Quaternion(Precision * matrix4, Precision* q){
     Precision a[9];
     Matrix4x4_2_Matrix3x3(matrix4, a);
     
     Precision trace = a[0] + a[4] + a[8]; // I removed + 1.0f; see discussion with Ethan
     if( trace > 0 ) {// I changed M_EPSILON to 0
       Precision s = 0.5f / sqrtf(trace+ 1.0f);
       q[0] = 0.25f / s;
       q[1] = ( a[7] - a[5] ) * s;
       q[2] = ( a[2] - a[6] ) * s;
       q[3] = ( a[3] - a[1] ) * s;
     } 
     else {
       if ( a[0] > a[4] && a[0] > a[8] ) {
	 Precision s = 2.0f * sqrtf( 1.0f + a[0] - a[4] - a[8]);
	 q[0] = (a[7] - a[5] ) / s;
	 q[1] = 0.25f * s;
	 q[2] = (a[1] + a[3] ) / s;
	 q[3] = (a[2] + a[6] ) / s;
       } else if (a[4] > a[8]) {
	 Precision s = 2.0f * sqrtf( 1.0f + a[4] - a[0] - a[8]);
	 q[0] = (a[2] - a[6] ) / s;
	 q[1] = (a[1] + a[3] ) / s;
	 q[2] = 0.25f * s;
	 q[3] = (a[5] + a[7] ) / s;
       } else {
	 Precision s = 2.0f * sqrtf( 1.0f + a[8] - a[0] - a[4] );
	 q[0] = (a[3] - a[1] ) / s;
	 q[1] = (a[2] + a[6] ) / s;
	 q[2] = (a[5] + a[7] ) / s;
	 q[3] = 0.25f * s;
       }
     }



     //precision T = M[0]+M[4]+M[8]+1;
     //if (T>0){
     //  precision R = sqrt(t);
     //  precision S = 1/(2*R);
     //  quaternion[0] = R/2;
     //  quaternion[1] = (M[7] - M[5])*S; 
     //  quaternion[2] = (M[2] - M[6])*S; 
     //  quaternion[3] = (M[3] - M[1])*S; 
     //  return 0; 
     //}
     //
     ////Determine biggest diagonal element
     //Precision max=0.0;
     //int argmax = -1;
     //for(int i=0; i<3; i++){
     //  Precision q =  fabs(M[i*3+i]);
     //  if (q>max){
     //	 max = q;
     //	 argmax = i;
     //  }
     //}
     //int u,v,w;
     //u = argmax;
     //v = (argmax+1)%3;
     //w = (argmax+2)%3;
     //
     //Precision a = 1+M[u*3+u]-M[v*3+v]-M[w*3+w];
     //if(a >0){
     //  Precision r = sqrt(a);
     //  Precision s = 1/(2*r);
     //  quaternion[0] = (M[w*3+v] - M[v*3+w])*s; 
     //  quaternion[1] = r/2;		    
     //  quaternion[2] = (M[u*3+v] + M[v*3+u])*s; 
     //  quaternion[3] = (M[w*3+u] + M[u*3+w])*s; 
     //}
     //else{
     //  fprintf(stderr, "Quaternion impossible to compute? \n");       
     //  exit(1);
     //}
     //
     //return argmax;
   }


      
   // only work for unit quaternion (need scaling otherwise)
   static void UnitQuaternion_2_Matrix3x3(Precision* quaternion, Precision* matrix3, int offset=0){
     Precision w = quaternion[0];
     Precision x = quaternion[1];
     Precision y = quaternion[2];
     Precision z = quaternion[3];
     Precision ww = w*w;
     Precision wx = w*x;
     Precision wy = w*y;
     Precision wz = w*z;
     Precision xx = x*x;
     Precision yy = y*y;
     Precision zz = z*z;
     Precision xy = x*y;
     Precision xz = x*z;
     Precision yz = y*z;

     matrix3[(0+offset)%3*3+(0+offset)%3] = ww + xx - yy - zz; 
     matrix3[(1+offset)%3*3+(0+offset)%3] = 2.0*(wz + xy);
     matrix3[(2+offset)%3*3+(0+offset)%3] = 2.0*(-wy + xz);
     matrix3[(0+offset)%3*3+(1+offset)%3] = 2.0*(-wz + xy);  
     matrix3[(1+offset)%3*3+(1+offset)%3] = ww - xx + yy - zz;
     matrix3[(2+offset)%3*3+(1+offset)%3] = 2.0*(wx + yz);
     matrix3[(0+offset)%3*3+(2+offset)%3] = 2.0*(wy + xz);
     matrix3[(1+offset)%3*3+(2+offset)%3] = 2.0*(-wx + yz);
     matrix3[(2+offset)%3*3+(2+offset)%3] = ww - xx - yy + zz;
   }

   // only write in the upper-left 3x3 matrix of the 4x4 matrix
   static void UnitQuaternion_2_Matrix4x4(Precision* quaternion, Precision* matrix4, int offset=0){
     Precision w = quaternion[0];
     Precision x = quaternion[1];
     Precision y = quaternion[2];
     Precision z = quaternion[3];
     Precision ww = w*w;
     Precision wx = w*x;
     Precision wy = w*y;
     Precision wz = w*z;
     Precision xx = x*x;
     Precision yy = y*y;
     Precision zz = z*z;
     Precision xy = x*y;
     Precision xz = x*z;
     Precision yz = y*z;

     matrix4[(0+offset)%3*4+(0+offset)%3] = ww + xx - yy - zz; 
     matrix4[(1+offset)%3*4+(0+offset)%3] = 2.0*(wz + xy);
     matrix4[(2+offset)%3*4+(0+offset)%3] = 2.0*(-wy + xz);
     matrix4[(0+offset)%3*4+(1+offset)%3] = 2.0*(-wz + xy);  
     matrix4[(1+offset)%3*4+(1+offset)%3] = ww - xx + yy - zz;
     matrix4[(2+offset)%3*4+(1+offset)%3] = 2.0*(wx + yz);
     matrix4[(0+offset)%3*4+(2+offset)%3] = 2.0*(wy + xz);
     matrix4[(1+offset)%3*4+(2+offset)%3] = 2.0*(-wx + yz);
     matrix4[(2+offset)%3*4+(2+offset)%3] = ww - xx - yy + zz;
   }




   static inline void MultiplyQuaternion(Precision *qout, Precision* qleft, Precision* qr){
     Precision ql[4];
     Copy_4(qleft, ql);
     
     qout[0] =  ql[0]*qr[0] - ql[3]*qr[1] + ql[2]*qr[2] + ql[1]*qr[3];
     qout[1] =  ql[3]*qr[0] + ql[0]*qr[1] - ql[1]*qr[2] + ql[2]*qr[3];
     qout[2] = -ql[2]*qr[0] + ql[1]*qr[1] + ql[0]*qr[2] + ql[3]*qr[3];
     qout[3] = -ql[1]*qr[0] - ql[2]*qr[1] - ql[3]*qr[2] + ql[0]*qr[3];
   }
     
   static inline void Quaternion2EulerAngles(Precision * q, Vector3 out){
     out[0] = atan2(2*(q[0]*q[1]+q[2]*q[3]), 1-2*(q[1]*q[1]+q[2]*q[2]));
     out[1] = asin(2*(q[0]*q[2] - q[3]*q[1]));
     out[2] = atan2(2*(q[0]*q[3]+q[1]*q[2]), 1-2*(q[2]*q[2]+q[3]*q[3]));
   }

   //// euler angle to matrix4
  //matrix[0] = cos(b)*cos(c); 
  //matrix[1] = -cos(b)*sin(c);
  //matrix[2] = -sin(b);
  //matrix[4] = sin(a)*sin(b)*cos(c)+cos(a)*sin(c);
  //matrix[5] = -sin(a)*sin(b)*sin(c)+cos(a)*cos(c);
  //matrix[6] =  sin(a)*cos(b);
  //matrix[8] = cos(a)*sin(b)*cos(c)-sin(a)*sin(c);
  //matrix[9] = -cos(a)*sin(b)*sin(c)-sin(a)*cos(c);
  //matrix[10]=  cos(a)*cos(b);

   // determinant of a transformation matrix
   // as the last line is [0,0,0,1] we compute the determinant of the 3x3 matrix
   static inline Precision DetTransformMatrix_4(Precision* in){ // 4x4 matrix
     Precision a[3],b[3],c[3];
     // Det(a^t) = Det(a)
     Copy_3(in+0,a);
     Copy_3(in+4,b);
     Copy_3(in+8,c);
     return Det_3(a,b,c);
   }







   static inline int IntersectionLinePlane(Precision* Oline, Precision* Dline, 
					   Precision * Oplane, Precision* Nplane,
					   Precision* point){
     Precision S[3];
     Substract_3(S,Oline, Oplane);
     
     Precision     D = DotProduct_3(Nplane, Dline);
     Precision     N = -DotProduct_3(Nplane, S);
     
     if ( (D >= 0 && D < EPSILON) || (D < 0 && D > -EPSILON) ) {          // segment is parallel to plane
       if (N == 0)                     // segment lies in plane
	 return 2;
       else
	 return 0;                   // no intersection
     }
     
     // compute intersect param
     Precision sI = N / D;
     
     Mult_3(point,sI, Dline);
     Add_3(point, Oline);                 // compute segment intersect point
     return 1;
   }
   
   
   static inline int IntersectionPlanePlane(Precision* Oplane1, Precision* Nplane1, 
					    Precision* Oplane2, Precision* Nplane2,
					    Precision* Oline, Precision* Dline){
     CrossProduct_3(Dline,Nplane1,Nplane2);         // cross product
     Precision ax = (Dline[0] >= 0 ? Dline[0] : -Dline[0]);
     Precision ay = (Dline[1] >= 0 ? Dline[1] : -Dline[1]);
     Precision az = (Dline[2] >= 0 ? Dline[2] : -Dline[2]);

     // test if the two planes are parallel
     if ((ax+ay+az) < EPSILON) {       // Pn1 and Pn2 are near parallel
       // test if disjoint or coincide
       Precision v[3]; 
       Substract_3(v, Oplane1, Oplane2);
       if (DotProduct_3(Nplane1, v) == 0)         // Pn2.V0 lies in Pn1
	 return 1;                   // Pn1 and Pn2 coincide
       else
	 return 0;                   // Pn1 and Pn2 are disjoint
     }

     // Pn1 and Pn2 intersect in a line
     // first determine max abs coordinate of cross product
     int      maxc;                      // max coordinate
     if (ax > ay) {
       if (ax > az)
	 maxc = 1;
       else maxc = 3;
     }
     else {
       if (ay > az)
	 maxc = 2;
       else maxc = 3;
     }

     // next, to get a point on the intersect line
     // zero the max coord, and solve for the other two
     Precision d1, d2;           // the constants in the 2 plane equations
     d1 = -DotProduct_3(Nplane1, Oplane1);  
     d2 = -DotProduct_3(Nplane2, Oplane2);  
    
     switch (maxc) {            // select max coordinate
     case 1:                    // intersect with x=0
       Oline[0] = 0;
       Oline[1] = (d2*Nplane1[2] - d1*Nplane2[2]) / Dline[0];
       Oline[2] = (d1*Nplane2[1] - d2*Nplane1[1]) / Dline[0];
       break;
     case 2:                    // intersect with y=0
       Oline[0] = (d1*Nplane2[2] - d2*Nplane1[2]) / Dline[1];
       Oline[1] = 0;
       Oline[2] = (d2*Nplane1[0] - d1*Nplane2[0]) / Dline[1];
       break;
     case 3:                    // intersect with z=0
       Oline[0] = (d2*Nplane1[1] - d1*Nplane2[1]) / Dline[2];
       Oline[1] = (d1*Nplane2[0] - d2*Nplane1[0]) / Dline[2];
       Oline[2] = 0;
     }
    
     return 2;
   }

   static inline void GetCircle_3(Precision* a,Precision* b,Precision* c, Precision* center, Precision& radius){ // 4x4 matrix
     //cc test if 3points aligned 
     Precision n[3];
     TriangleNormal_3(n,a,b,c);     
     Precision Np1[3],Np2[3],Op1[3],Op2[3];
     
     MiddlePoint_3(Op1,a,0.5,b);
     MiddlePoint_3(Op2,a,0.5,c);
     Substract_3(Np1,b,a);
     Substract_3(Np2,c,a);
     Normalize_3(Np1);
     Normalize_3(Np2);
   
     int ret = 0;
     Precision Ol[3],Dl[3];
     ret = IntersectionPlanePlane(Op1,Np1, Op2,Np2, Ol,Dl);
     
     ret =  IntersectionLinePlane(Ol,Dl,a,n,center);
     
     radius= Distance_3(center,a);
   }
   

   // reorder the input pointer so that they are ordered from the closest to the centroid to the most remote
   static inline void ReorderByDistanceToCentroid(std::vector<std::pair<Precision*, int> >& V){ // 4x4 matrix
     Precision center[3] = {0.0,0.0,0.0};
     for(unsigned int i = 0; i<V.size(); i++){
       Add_3(center,V[i].first);
     }
     Divide_3(center,(double)V.size());


     std::vector<std::pair<Precision, std::pair<Precision*,int> > > D;
     for(unsigned int i = 0; i<V.size(); i++){
       D.push_back(std::pair<Precision, std::pair<Precision*,int> > (Distance_3(center,V[i].first),V[i]));
     }
     sort(D.begin(), D.end());
     V.clear();
     for(unsigned int i = 0; i<D.size(); i++){
       V.push_back(D[i].second);
     }
   }



};





#endif
