//
// Copyright (C) 2010  Clement Creusot
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "mesh.h"
#include "geom.h"
#include <algorithm> // for find_if
#include "common.h"
#include <limits>



#ifdef USE_VTK

Mesh::Mesh(vtkPolyData * pd, bool mCheck){
  this->polyData = pd;
  this->manifoldCheck = mCheck;  
  this->Init();

  int npoints = pd->GetNumberOfPoints();
  int ncells = pd->GetNumberOfCells();  
  this->hEdges.resize(npoints);
  this->faces.resize(ncells);
  this->points.resize(npoints);
  for(int i=0; i!=(int)this->points.size(); i++){
    this->points[i] = new double[3];
  }  

  this->BuildMeshFromPolyData(pd);
  this->InitPostProc();
}


// Build a table that contain the list of neighboor for each vertex
void Mesh::BuildMeshFromPolyData(vtkPolyData *pd)
{
  int i;  
  int npoints = pd->GetNumberOfPoints();
  int ncells = pd->GetNumberOfCells();
  
  // Copy the points coordinates 
  vtkPoints * vtkCoords = pd->GetPoints();   
  for (i=0; i<npoints; i++){
    this->points[i][0] = vtkCoords->GetPoint(i)[0];   
    this->points[i][1] = vtkCoords->GetPoint(i)[1];   
    this->points[i][2] = vtkCoords->GetPoint(i)[2];   

    this->boundmin[0] = std::min(boundmin[0],points[i][0]);
    this->boundmin[1] = std::min(boundmin[1],points[i][1]);
    this->boundmin[2] = std::min(boundmin[2],points[i][2]);
    this->boundmax[0] = std::max(boundmax[0],points[i][0]);
    this->boundmax[1] = std::max(boundmax[1],points[i][1]);
    this->boundmax[2] = std::max(boundmax[2],points[i][2]);
  }

  // Copy faces and fill HalfEdge structure
  for (i = 0; i < ncells; i++){
    vtkIdType ctype = pd->GetCellType(i);
    
    // Only accept triangles 
    if (ctype == VTK_TRIANGLE){
      vtkIdType *pts;
      vtkIdType npts;
      pd->GetCellPoints (i, npts, pts);
      int u = this->faces[i].vertex[0] = pts[0];
      int v = this->faces[i].vertex[1] = pts[1];
      int w = this->faces[i].vertex[2] = pts[2];
  
      // the face number is associated with each
      // oriented halfedge x->y
      this->faces[i].hEdge[0] = this->addHalfEdge(u, v, i);
      this->faces[i].hEdge[1] = this->addHalfEdge(v, w, i);
      this->faces[i].hEdge[2] = this->addHalfEdge(w, u, i);

      this->faces[i].hEdge[0]->next = this->faces[i].hEdge[1];
      this->faces[i].hEdge[1]->next = this->faces[i].hEdge[2];     
      this->faces[i].hEdge[2]->next = this->faces[i].hEdge[0];
    }
    else{
      PRINT_WARNING("This vtkPolyData contains cells that are not triangles;");
      PRINT_WARNING("Note that those faces will not be taken into account.");
    }
  }
}

#endif



Mesh::Mesh(const char *filename, bool mCheck){
  this->manifoldCheck = mCheck;  
  this->Init();
  this->BuildMeshFromFile(filename);
  this->InitPostProc();
}


void Mesh::Init(){
  this->boundmin[0] = std::numeric_limits<double>::max();
  this->boundmin[1] = std::numeric_limits<double>::max();
  this->boundmin[2] = std::numeric_limits<double>::max();
  this->boundmax[0] = -std::numeric_limits<double>::max();
  this->boundmax[1] = -std::numeric_limits<double>::max();
  this->boundmax[2] = -std::numeric_limits<double>::max();
}

void Mesh::InitPostProc(){
  this->Check();
  this->CreateBorderList();

}


// Construct the mesh by reading an .obj file.
void Mesh::BuildMeshFromFile(const char *filename){  
  std::ifstream in;
  std::string buff, name;
  if (!filename) {
    PRINT_ERROR("Empty filename");
  }
    
  in.open(filename);
  if (!in) {
    PRINT_ERROR("File can't be opened");
  }

  // Set initial size to 10000.  
  int points_vector_size = 10000;
  this->points.resize(points_vector_size);
  int faces_vector_size = 10000;
  this->faces.resize(faces_vector_size);

  int read_vertex_count = 0;
  int read_face_count = 0;
  while( !in.eof() ){
    if (buff[0] == '#'){
      // if we read a comment, read all the line
      getline(in, buff);
    }
    else if (buff[0] == 'v'){
      double x,y,z;
      in >> x;
      in >> y;
      in >> z;
      // Double the destination size if too small.
      if( read_vertex_count > points_vector_size -1){
	points_vector_size *= 2;
	this->points.resize(points_vector_size);	
      } 
      this->points[read_vertex_count] = new double[3];
      this->points[read_vertex_count][0] = x;
      this->points[read_vertex_count][1] = y;
      this->points[read_vertex_count][2] = z;
      
      this->boundmin[0] = std::min(boundmin[0],x);
      this->boundmin[1] = std::min(boundmin[1],y);
      this->boundmin[2] = std::min(boundmin[2],z);
      this->boundmax[0] = std::max(boundmax[0],x);
      this->boundmax[1] = std::max(boundmax[1],y);
      this->boundmax[2] = std::max(boundmax[2],z);
      read_vertex_count++;
    }
    else if (buff[0] == 'f'){
      if ( read_face_count == 0 ){
	// Reduce sizes to the bare minimum.
	this->points.resize(read_vertex_count);	  
	this->hEdges.resize(read_vertex_count);	  	
      }
      int u[3];
      for ( int k=0; k<3; k++ ){	
	in >> buff;
	size_t cutAt = buff.find_first_of("/"); 
	if ( ( cutAt != buff.npos ) && cutAt >= 0){
	  std::istringstream ss(buff.substr(0,cutAt));
	  ss >> u[k];
	} else {
	  std::istringstream ss(buff);
	  ss >> u[k];
	}
	u[k]--; // Obj index starts at 1 not 0.
      }
      // Double the destination size if too small.
      if ( read_face_count > faces_vector_size-1 ){
	faces_vector_size *= 2;
	this->faces.resize(faces_vector_size);	
      } 
      int i = read_face_count;
      // The face number is associated with each
      // oriented halfedge x->y.
      this->faces[i].vertex[0] = u[0];
      this->faces[i].vertex[1] = u[1];
      this->faces[i].vertex[2] = u[2];
      this->faces[i].hEdge[0] = this->addHalfEdge(u[0], u[1], i);
      this->faces[i].hEdge[1] = this->addHalfEdge(u[1], u[2], i);
      this->faces[i].hEdge[2] = this->addHalfEdge(u[2], u[0], i);
      this->faces[i].hEdge[0]->next = this->faces[i].hEdge[1];
      this->faces[i].hEdge[1]->next = this->faces[i].hEdge[2];     
      this->faces[i].hEdge[2]->next = this->faces[i].hEdge[0];
      read_face_count++;    
    }
    in >>buff;
  }
  in.close();

  // Reduce size to the bare minimum.
  this->faces.resize(read_face_count);	  

}


// Construct the mesh by reading an .obj file.
void Mesh::Export(const char * filename){  
  std::ofstream out;
  out.open(filename, std::fstream::out);
  if(!out) { 
    PRINT_ERROR("Unable to open the file %s.",filename);
  }
  for(unsigned int j=0; j<this->points.size(); j++){
    out << "v "
	<< this->points[j][0] << " "      
	<< this->points[j][1] << " "     
	<< this->points[j][2] << std::endl;      
  }
  for(unsigned int j=0; j<this->faces.size(); j++){
    out << "f "
	<< this->faces[j].vertex[0]+1 << " "      
	<< this->faces[j].vertex[1]+1 << " "     
	<< this->faces[j].vertex[2]+1 << std::endl;      
  }
  out.close();    
}


Mesh::~Mesh(){  
  this->hEdges.clear();
  this->faces.clear();
  
  std::vector<double*>::iterator 
    itb = this->points.begin(),
    ite = this->points.end();  
  for(; itb != ite; itb++){
    delete [] (*itb);
  }  
  this->points.clear();

}



HalfEdge * Mesh::addHalfEdge(int u, int v, int face){
  
  std::list<HalfEdge>::iterator loc1,loc2;
  loc1 = std::find_if(hEdges[u].begin(), hEdges[u].end(), HEdge_Finder(v));
  if (loc1 == hEdges[u].end()){
    // if the halfedge doesnt exist: create it.
    HalfEdge he1;
    he1.vertex = v;
    he1.face = face;
    // add the halfedge
    hEdges[u].push_back(he1);
    HalfEdge& ref = hEdges[u].back();

    // test if the opposite half edge exist
    loc2 = std::find_if(hEdges[v].begin(), hEdges[v].end(), HEdge_Finder(u));
    if (loc2 == hEdges[v].end()){
      // if it doesnt exist create it and
      // associate it with no face (halfedge on border)
      HalfEdge he2;
      he2.vertex = u;
      he2.face = -1; //on border
      he2.opposite = &ref; //pointer to he1
      hEdges[v].push_back(he2);
      // pointer to the opposite halfedge
      ref.opposite = &hEdges[v].back(); 
    }
    else{
      PRINT_ERROR("An half edge was present without is opposite halfedge");
    }
    
    return &ref;
  }
  else{
    // test if the opposite half edge exist
    loc2 = std::find_if(hEdges[v].begin(), hEdges[v].end(), HEdge_Finder(u));
    if (loc2 == hEdges[v].end()){
      PRINT_ERROR("An halfedge have been created without its opposite halfedge");
    }
    if (&(*loc1) != ((*loc2).opposite)){
      PRINT_ERROR("The opposite halfedge is not unique");
    }
    if (this->manifoldCheck){
      if ((*loc1).face != -1){
	printf("%d %d %d\n",(*loc1).face, (*loc1).vertex, (*loc1).opposite->vertex);  
	PRINT_ERROR("This halfedge related to this face have been added more than once");
      }
    }
    // This halfedge is no longer on border as we have found it while visiting a the current face 
    (*loc1).face = face;

    return &(*loc1);  
  }  
}






void Mesh::Print(){
  printf("Mesh Details:\n");
  printf("=============\n");
  printf("Point Nb: %d\n", (int)points.size());
  printf("Face  Nb: %d\n", (int)faces.size());

  for(int i=0; i< (int)faces.size(); i++){
    printf("face %d : ", i);
    printf("%d %d %d,\n", faces[i].vertex[0],faces[i].vertex[1],faces[i].vertex[2]);
    printf("  v %d %d %d,\n", faces[i].hEdge[0]->vertex,faces[i].hEdge[1]->vertex,faces[i].hEdge[2]->vertex);
    printf("  f %d %d %d,\n", faces[i].hEdge[0]->face,faces[i].hEdge[1]->face,faces[i].hEdge[2]->face);
  }

  int hedgeNb =0;
  for(int i=0; i< (int)hEdges.size(); i++){
    std::list<HalfEdge>::iterator 
      iEdgeb = this->hEdges[i].begin(),
      iEdgee = this->hEdges[i].end();  
    printf("edges %d ->: \n", i);
    for (; iEdgeb!=iEdgee; iEdgeb++){
      hedgeNb++;
      printf("  %d o%d f%d,\n", (*iEdgeb).vertex, (*iEdgeb).opposite->vertex, (*iEdgeb).face);
    }
  }
  printf("HEdge  Nb: %d\n", hedgeNb);

}

void Mesh::CreateBorderList(){
  if (borderHalfEdges.size() !=0){
    return;
  }
  for(int i=0; i< (int)hEdges.size(); i++){
    std::list<HalfEdge>::iterator 
      iEdgeb = this->hEdges[i].begin(),
      iEdgee = this->hEdges[i].end();  
    for (; iEdgeb!=iEdgee; iEdgeb++){
      if((*iEdgeb).face == -1){
	std::vector<HalfEdge*> bords;
	bords.push_back(&(*iEdgeb));
	// find an border starting in x
	int x = (*iEdgeb).vertex;
	std::list<HalfEdge>::iterator 
	  itb = this->hEdges[x].begin(),
	  ite = this->hEdges[x].end();
	for(;itb!=ite; itb++){
	  if ((*itb).face == -1){
	    bords.push_back(&(*itb));	    
	    break; //cc only one for now
	  }
	}
	borderHalfEdges.push_back(bords);
      }
    }
  }
}

void Mesh::Check(){
  bool allOk = true;
  for(int f=0; f< (int)faces.size(); f++){
    if (faces[f].hEdge[0]->next != faces[f].hEdge[1] ||
	faces[f].hEdge[1]->next != faces[f].hEdge[2] ||
	faces[f].hEdge[2]->next != faces[f].hEdge[0] ) {
      PRINT_ERROR("The 'next' pointers are not correct.");
    }
    allOk = allOk & (faces[f].hEdge[0]!=NULL); 
    allOk = allOk & (faces[f].hEdge[1]!=NULL);
    allOk = allOk & (faces[f].hEdge[2]!=NULL);
    allOk = allOk & (faces[f].hEdge[0]!=NULL); 
    allOk = allOk & (faces[f].hEdge[1]!=NULL);
    allOk = allOk & (faces[f].hEdge[2]!=NULL);


    for(int j=0; j<3; j++){
      int k = (j+1)/3;// the last edge is x=..[2] y=..[0]
      int x = faces[f].vertex[j];
      int y = faces[f].vertex[j+1 -3*k];
      std::list<HalfEdge>::iterator loc1, loc2;
      loc1 = std::find_if(hEdges[x].begin(), hEdges[x].end(), HEdge_Finder(y));
      if (loc1 == hEdges[x].end())    
	PRINT_ERROR("An halfedge is missing.");
      if (&(*loc1) != faces[f].hEdge[j])
	PRINT_ERROR("One of the hedge attached to the face dont match the hedge find is the mesh.");
      if ((*loc1).face != f){
	PRINT_ERROR("The face stored in the halfedge do not correspond.");
      }
      loc2 = std::find_if(hEdges[y].begin(), hEdges[y].end(), HEdge_Finder(x));
      if (loc2 == hEdges[y].end())    
	PRINT_ERROR("The opposite halfedge is missing.");
      if (&(*loc1) != (*loc2).opposite)
	PRINT_ERROR("The two halfedges are not attached.");
      if (&(*loc2) != (*loc1).opposite)
	PRINT_ERROR("The two halfedges are not attached.");
    }	
  }
  

  int hEdgeNb =0;
  int borderHEdgeNb =0;
  for(int i=0; i< (int)hEdges.size(); i++){
    std::list<HalfEdge>::iterator 
      iEdgeb = this->hEdges[i].begin(),
      iEdgee = this->hEdges[i].end();  
    for (; iEdgeb!=iEdgee; iEdgeb++){
      if((*iEdgeb).face != -1){
	allOk = allOk & ((*iEdgeb).next != NULL);
      }
      hEdgeNb++;
      if ( (*iEdgeb).face == -1 )
	borderHEdgeNb++;
    }
  }

  if (not allOk){
    PRINT_ERROR("Some pointers are null.");
  }
  
  if (hEdgeNb - borderHEdgeNb !=  3*(int)faces.size()){
    PRINT_ERROR("The number of half edge doesnt match the number of face x3.");
  }

}

void Mesh::ComputeBalloonImage(const int id,
			       const int radiusBinNb, 
			       const double radiusStepSize,
			       HMatrix& histogram) const{
  
  int rn = radiusBinNb;
  histogram.SetSize(1,rn,0.0);
  
  double O[3];
  Geom<double>::Copy_3(points[id],O);
  
  for (unsigned int i=0; i<points.size(); i++){
    double OP[3];
    Geom<double>::Substract_3(OP, points[i], O); 
    double d = Geom<double>::VectorSize_3(OP);
    int a = int(d/radiusStepSize);
    if (a < radiusBinNb) {
      histogram[0][a] += 1.0;
    }    
  }
}


void Mesh::ComputeSpinImage(const int id, const double* normal, 
			    const int radiusBinNb, 
			    const int heightBinNbOver, const int heightBinNbUnder, 
			    const double radiusStepSize, const double heightStepSize, 
			    HMatrix& histogram) const{
  
  int hn = heightBinNbOver + heightBinNbUnder;
  int rn = radiusBinNb;
  histogram.SetSize(rn,hn,0.0);
  
  double O[3];
  Geom<double>::Copy_3(points[id],O);
  

  for (unsigned int i=0; i<points.size(); i++){
    double OP[3], OP_r[3], OP_h[3];
    Geom<double>::Substract_3(OP, points[i], O); 
    double a = Geom<double>::DotProduct_3(OP,normal);
    Geom<double>::Mult_3(OP_h,a,normal);
    Geom<double>::Substract_3(OP_r,OP,OP_h);
    double b = Geom<double>::VectorSize_3(OP_r);
    // a along m, b along n
    
    int i_h = int (a/heightStepSize) + heightBinNbUnder;
    int i_r = int(b/radiusStepSize);
    if (i_r>=0 && i_r<rn && i_h>=0 && i_h<hn) {
      histogram[i_r][i_h] += 1.0;
    }
    
  }
  
}


void Mesh::ComputeSpinImageAB(const int a,  const int b,
			      const int radiusBinNb, 
			      const int heightBinNb, 
			      const double radiusStepSize,
			      HMatrix& histogram) const{
  
  int hn = heightBinNb;
  int rn = radiusBinNb;
  histogram.SetSize(rn,hn,0.0);
  
  double A[3],B[3],ABn[3];
  Geom<double>::Copy_3(points[a],A);
  Geom<double>::Copy_3(points[b],B);
  Geom<double>::Substract_3(ABn, B, A); 
  double distAB = Geom<double>::VectorSize_3(ABn);
  Geom<double>::Normalize_3(ABn); 
  double heightStepSize = distAB/(double)heightBinNb;  

  for (unsigned int i=0; i<points.size(); i++){
    double AP[3], AP_r[3], AP_h[3];
    Geom<double>::Substract_3(AP, points[i], A); 
    double y = Geom<double>::DotProduct_3(AP,ABn);
    Geom<double>::Mult_3(AP_h,y,ABn);
    Geom<double>::Substract_3(AP_r,AP,AP_h);
    double x = Geom<double>::VectorSize_3(AP_r);
    
    int i_h = int (y/heightStepSize);
    int i_r = int(x/radiusStepSize);
    if (i_r>=0 && i_r<rn && i_h>=0 && i_h<hn) {
      histogram[i_r][i_h] += 1.0;
    }
    
  }
  
}


void Mesh::ComputePairwiseGeometricHistogram(const int id, const std::vector<double*>& normals, 
					     const int angleBinNb, 
					     const int heightBinNbOver, const int heightBinNbUnder, 
					     const double heightStepSize, 
					     const double maxRadius,
					     HMatrix& histogram) const{
  
  int hn = heightBinNbOver + heightBinNbUnder;
  int rn = angleBinNb;
  double angleStep = PI/double(rn);
  histogram.SetSize(rn,hn,0.0);
  
  double O[3];
  double N[3];
  Geom<double>::Copy_3(points[id],O);
  Geom<double>::Copy_3(normals[id],N);
  
  for (unsigned int i=0; i<points.size(); i++){
    double OP[3];//, OP_r[3], OP_h[3];
    Geom<double>::Substract_3(OP, points[i], O); 
    double a = Geom<double>::DotProduct_3(OP,N);
    double d = Geom<double>::VectorSize_3(OP);
    double b = acos(Geom<double>::DotProduct_3(normals[i],N));

    // a along m, b along n    
    int i_h = int (a/heightStepSize) + heightBinNbUnder;
    int i_r = int(b/angleStep);
    if (d<maxRadius && i_r>=0 && i_r<rn && i_h>=0 && i_h<hn) {
      histogram[i_r][i_h] += 1.0;
    }
    
  }
  
}



// for each vertex make the mean of the normal of connected faces
void Mesh::MakeNormals(std::vector<double*>& normals) const{
  int i, j;
  int npoint = (int)this->Points().size();
  double dsizeUV=0.0, dsizeUW=0.0, dsizeVW=0.0;
  double weight1=0.0, weight2=0.0, weight3=0.0;
  double uv[3], uw[3], vw[3];   
  double vect1[3], vect2[3], vect3[3];   

  // Initialisation 
  normals.resize(npoint);
  for(i=0; i<npoint; i++){    
    normals[i] = new double[3];
    normals[i][0] = 0.0;
    normals[i][1] = 0.0;
    normals[i][2] = 0.0;
  }

  // For all the triangles, add the normal of the face
  // to the normal of the three vertices
  for (j = 0; j < (int)this->Faces().size(); j++){
    
    const int * vertex = this->Faces()[j].vertex;
    int u = vertex[0];
    int v = vertex[1];
    int w = vertex[2];
    
    const std::vector<double*>& points = this->Points();

    // FIXME : simplify this function

    // the first vector
    Geom<double>::Substract_3(uw, points[u], points[w]);
    dsizeUW=Geom<double>::VectorSize_3(uw);
    if(dsizeUW==0.0)dsizeUW=1.0;

    // the second vector
    Geom<double>::Substract_3(uv, points[u], points[v]);
    dsizeUV=Geom<double>::VectorSize_3(uv);
    if(dsizeUV==0.0)dsizeUV=1.0;

    // the third vector
    Geom<double>::Substract_3(vw, points[v], points[w]);
    dsizeVW=Geom<double>::VectorSize_3(vw);
    if(dsizeVW==0.0)dsizeVW=1.0;

      
    // determine the non normalized normals for the face
    Geom<double>::CrossProduct_3(vect1,uv,uw);
    Geom<double>::CrossProduct_3(vect2,uv,vw);
    Geom<double>::CrossProduct_3(vect3,uw,vw);

      
    // ponderate by something proportional to the area
    //weight1 = dsizeUV*dsizeUW;
    //weight2 = dsizeUV*dsizeVW;
    //weight3 = dsizeUW*dsizeVW;
    // Other ponderation (Nelson Max, Graphical Tools 4(2))
    weight1 = 1.0/(dsizeUV*dsizeUV*dsizeUW*dsizeUW);
    weight2 = 1.0/(dsizeUV*dsizeUV*dsizeVW*dsizeVW);
    weight3 = 1.0/(dsizeUW*dsizeUW*dsizeVW*dsizeVW);
	 
    // add to the normal
    normals[u][0] += weight1*vect1[0];
    normals[u][1] += weight1*vect1[1];
    normals[u][2] += weight1*vect1[2];

    normals[v][0] += weight2*vect2[0];
    normals[v][1] += weight2*vect2[1];
    normals[v][2] += weight2*vect2[2];

    normals[w][0] += weight3*vect3[0];
    normals[w][1] += weight3*vect3[1];
    normals[w][2] += weight3*vect3[2];    
  }
  
  //for each border node
  // take the main direction and the two face on border normals
  // define symetry plane
  // add to normal the symetric of all connected faces

  // normalisation of the normals
  for(i=0; i<npoint; i++){    
    Geom<double>::Normalize_3(normals[i]);
  }

}


int Mesh::GetObjMeshSize(std::string objFilename){
  std::ifstream in;
  std::string buff;
  int vertexNb = 0;
  
  in.open(objFilename.c_str());
  if(!in) { 
    PRINT_ERROR("Unable to open the file %s",objFilename.c_str());
  }  
  while (std::getline(in, buff)) {
    if (buff.size() == 0){
      continue;
    }
    std::istringstream iss(buff);
    char c;
    iss >> c;
    if (c == '#'){
      // if we read a comment, read all the line
      // do nothing
    }
    else if (c == 'v'){
      vertexNb ++;
    }
  }

  return vertexNb;
}
