//
// Copyright (C) 2010  Clement Creusot
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef MESHH
#define MESHH

#ifdef USE_VTK
#include "vtkPolyData.h"
#endif

#include <list>
#include <vector>
#include "histogram.h"


class HalfEdge
{
 public:
  int vertex; //vertex at the end of the half edge
  int face; //-1 means that the halfedge is on the border
  HalfEdge * opposite;
  HalfEdge * next;
  bool IsBorderEdge()const {return (opposite->face==-1 || face ==-1 );}
  bool IsBorderHalfEdge()const {return face ==-1 ;}

  static bool Compare (HalfEdge* first, HalfEdge* second){
    if (first->vertex == second->vertex &&
	first->opposite->vertex == second->opposite->vertex &&
	first->face == second->face ) return true;
    else return false;
  }

};

// In order to find an halfedge using only the pointed vertex
struct HEdge_Finder
{
 HEdge_Finder(int id)
   : target_vertex(id){
  };
  bool operator()( HalfEdge const& test) const{
    return test.vertex == target_vertex;
  };
  int target_vertex;
};


struct Face
{
  int vertex[3];
  HalfEdge * hEdge[3];
};



class Mesh
{
 public :
  
#ifdef USE_VTK
  Mesh(vtkPolyData * pd,bool manifoldCheck=true);
#endif
  Mesh(const char* filename, bool manifoldCheck=true);
  ~Mesh();

  void Init();  
  void InitPostProc();  
  void Print();  
  void Check();
  void MakeNormals(std::vector<double*>& normals) const;
  int FindClosestMeshID(const double* position)const;
  int FindClosestMeshIDBrut(const double* position)const;
  int FindClosestMeshID(const double* position, int initId) const;
  const int GetFaceNb() const {return faces.size();};
  const std::vector<std::list<HalfEdge> >& HEdges() const{return hEdges;}
  const std::vector<double *>& Points() const{return points;}
  const std::vector<Face>& Faces() const{return faces;}

  void ComputeBalloonImage(const int id,
			   const int radiusBinNb, 
			   const double radiusStepSize,
			   HMatrix& histogram) const;

  void ComputeSpinImage(const int id, const double* normal, const int n_side, 
			const int m_over, const int m_under, 
			const double n_length, const double m_length, 
			HMatrix& histogram) const;
  void ComputeSpinImageAB(const int a,  const int b,
			  const int radiusBinNb, 
			  const int heightBinNb, 
			  const double radiusStepSize,
			  HMatrix& histogram) const;
  void ComputePairwiseGeometricHistogram(const int id, const std::vector<double*>& normals, 
					 const int angleBinNb, 
					 const int heightBinNbOver, const int heightBinNbUnder, 
					 const double heightStepSize, 
					 const double maxRadius,
					 HMatrix& histogram) const;

  static int GetObjMeshSize(std::string objFilename);

#ifdef USE_VTK
  vtkPolyData* GetPolyData() const{return polyData;}
#endif

  void Export(const char *filename); 
  
 private:
  std::vector<std::list<HalfEdge> > hEdges;
  std::vector<Face> faces;
  std::vector<double *> points;

  double boundmin[3];
  double boundmax[3];
  bool manifoldCheck;

  // vector of pair <he_in, he_out> for each border point(idealy)
  // with he_in and he_out border halfedges
  // possibly several he_out -> pb
  std::vector<std::vector<HalfEdge*> > borderHalfEdges;


  void BuildMeshFromFile(const char *filename);
  HalfEdge* addHalfEdge(int u, int v, int face);
  void CreateBorderList();


#ifdef USE_VTK
  vtkPolyData* polyData;  
  void BuildMeshFromPolyData(vtkPolyData *pd);
#endif

};

#endif
