//
// Copyright (C) 2010  Clement Creusot
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "neighborhood.h"
#include "geom.h"
#include "common.h"
#include "stopwatch.h"

#include <algorithm> // for find_if(), sort

Neighborhood::Neighborhood(const Mesh* m, int n, const std::vector<int>*subset){
  this->mesh = m;
  if (n==0){
    PRINT_ERROR("Neighborhood of size 0");
  }
  this->maxLevel = n;
  this->distanceType = LEVEL_DISTANCE;
  BuildLargeNeighborhood(subset);
}

Neighborhood::Neighborhood(const Mesh* m, double d, const std::vector<int>*subset){
  this->mesh = m;
  if (d==0.0){
    PRINT_ERROR("Neighborhood of size 0");
  }
  this->maxDistance = d;
  this->distanceType = EUCLIDIAN_DISTANCE;
  this->BuildLargeNeighborhood(subset);
}


Neighborhood::~Neighborhood(){  
  neighborhoodBorder.clear();
  neighborhoodFaces.clear();
  neighborhoodPoints.clear();
}


inline bool Neighborhood::IsDistanceOk(int i, HalfEdge * he, int* level){
  if (this->distanceType == LEVEL_DISTANCE){
    return level[he->vertex] < this->maxLevel || 
      level[he->opposite->vertex] < this->maxLevel;

  }
  else if (this->distanceType == EUCLIDIAN_DISTANCE){
    double dist1 = Geom<double>::Distance_3(this->mesh->Points()[i],
					    this->mesh->Points()[he->vertex]); 
    double dist2 = Geom<double>::Distance_3(this->mesh->Points()[i],
					    this->mesh->Points()[he->opposite->vertex]); 
    return (dist1 <= this->maxDistance) && (dist2 <= this->maxDistance);
  }
  return false;
}

inline bool Neighborhood::IsDistanceOk(int i, int j, int* level){
  if (this->distanceType == LEVEL_DISTANCE){
    if(level!=NULL){
      return level[j] < this->maxLevel;
    }
    else{
      PRINT_ERROR("Level distance to be fixed");
    }
  }
  else if (this->distanceType == EUCLIDIAN_DISTANCE){
    double dist1 = Geom<double>::Distance_3(this->mesh->Points()[i],
					    this->mesh->Points()[j]); 
    return (dist1 <= this->maxDistance) ;
  }
  return false;
}





void Neighborhood::BuildLargeNeighborhood(const std::vector<int>*subset){
  int i;
  int nface  = (int)this->mesh->Faces().size();
  int npoint = (int)this->mesh->Points().size();
  int label[nface];
  int level[npoint];

  Stopwatch::Start("Neighborhood");

  // to save time we use a list for the construction that
  // we convert in vector for faster access
  std::vector< std::list<HalfEdge*> > neighborhoodBorder_tmp;
  neighborhoodBorder_tmp.resize(npoint);

  neighborhoodBorder.resize(npoint);
  neighborhoodFaces.resize(npoint);
  neighborhoodPoints.resize(npoint);
  
  // no need to initialise those structure
  // used as a bool map: (label[a] == i) <=> a visited
  // as i is unic, no need to reset.
  for(i=0; i<nface; i++){
    label[i] = -1;
  }
  for(i=0; i<npoint; i++){
    level[i] = 0;
  }

  //cc avoid that
  std::vector<int> selection ;
  if (subset==NULL){
    for(i=0; i<npoint; i++){
      selection.push_back(i);
    }
  }
  else{
    selection = (*subset);
  }  

  int emptyNeighborhoods = 0;
  std::vector<int>::iterator
    itab = selection.begin(),
    itae = selection.end();
  for(; itab!= itae; itab++){
    i = (*itab);
    std::list<HalfEdge*> meshBorderEdge;

    // initialisation: we add the border
    // define by the hedges that do not contain i
    // but which are on the faces touching i.
    /*  
      |  . .            |       __    
      | . i .    _\/_   |      /  \   
      |  . .      /\    |      \__/   
      |                 | 
      |   i    hedge(i) |  hedge(i).next (->border at t=0)
    */  
  
    std::list<HalfEdge>::const_iterator 
      itb = this->mesh->HEdges()[i].begin(), 
      ite = this->mesh->HEdges()[i].end();
    for(; itb != ite; itb++){
      if( ! (*itb).IsBorderHalfEdge() ){
	// if (i==50) printf("face %d\n", (*itb).face);
	neighborhoodBorder_tmp[i].push_back((*itb).next);
	neighborhoodFaces[i].push_back((*itb).face);
	label[(*itb).face] = i;
      }
      // there is no obligation to keep the border edges
      // however it can be usefull to know when a point is 
      // close to the border
      // Note that this additional border will change
      // the value of the volume (for example) on the border.
      if( (*itb).IsBorderEdge() ){
	if ((*itb).IsBorderHalfEdge()){
	  meshBorderEdge.push_back((*itb).opposite);
	}
	else{
	  meshBorderEdge.push_back((*itb).opposite->opposite);
	}      
      }
      //if (i==50) printf("poin %d\n", (*itb).vertex);
      if ( IsDistanceOk(i,(*itb).vertex, level)){
	level[(*itb).vertex] = 1;
	neighborhoodPoints[i].push_back((*itb).vertex);
      }
    }
    
    // we make the border grow until the level reach maxNeighborLevel
    // or some holes (model borders).
    unsigned int cntOutlier = neighborhoodBorder_tmp[i].size();
    while( !neighborhoodBorder_tmp[i].empty()){
      HalfEdge* he = neighborhoodBorder_tmp[i].front();
      if ( this->IsDistanceOk(i, he, level) ){
	if ( !he->IsBorderEdge() && label[he->opposite->face] != i){
	  // If the opposite halfedge is not on a neighbor
	  // face : add the face to the neighbor set
	  // remove the current halfedge from the border
	  // and add 0, 1 or 2 new halfedges on the border.
	  // we add 0 if the two other halfedge have their
	  // opposite on a treated face.
	  // we add 1 if only one halfedge touch a treated face
	  // else we add the two.
	  //   |           |                 |   
	  //  /|\  add 0  /|\      	     |
	  // /o| \       / | \   	     |
	  //   				     |
	  //                      ______     | 
	  //   /\    /\          /\    /\    |
	  //  / o\  /  \ add 1  /  \  /  \   | 
	  // /    \/    \      /    \/    \  |
	  //				     |
	  //                   /\  	     |
	  //	     add 2    /  \ 	     |
	  // _____	     /____\ 	     |
	  // \ o  /          \ o  / 	     |
	  //  \  /	      \  / 	     |
	  //   \/  	       \/            |
	  //
	  //if (i==50) printf("face %d\n", he->opposite->face);
	  bool nextExist = label[he->opposite->next->opposite->face] == i;
	  if ( !nextExist ){	      
	    neighborhoodBorder_tmp[i].push_back(he->opposite->next);
	  }
	  bool nextnextExist = (label[he->opposite->next->next->opposite->face] == i);
	  if ( !nextnextExist ){
	    neighborhoodBorder_tmp[i].push_back(he->opposite->next->next);
	  }
	  if( !nextExist && !nextnextExist){ // if add 2 hedges, also add the new vertex
	    //if (i==50) printf("poin %d\n", he->opposite->next->vertex);

	    if ( IsDistanceOk(i,he->opposite->next->vertex,level)){
	      neighborhoodPoints[i].push_back(he->opposite->next->vertex);
	      level[he->opposite->next->vertex] = 1+std::min(level[he->vertex], 
							     level[he->opposite->vertex]);
	    }
	  }

	  label[he->opposite->face] = i;	  
	  neighborhoodFaces[i].push_back(he->opposite->face);
	}
	// we add the mesh border edge
	else if(he->IsBorderEdge()){
	  meshBorderEdge.push_back(he);
	}
	cntOutlier = neighborhoodBorder_tmp[i].size();
      }
      else{
	neighborhoodBorder_tmp[i].push_back(he);
	cntOutlier--;
	if (cntOutlier == 0){
	  // we add the mesh border edge at the end
	  std::list<HalfEdge*>::iterator 
	    ibtb = neighborhoodBorder_tmp[i].begin(), 
	    ibte = neighborhoodBorder_tmp[i].end();
	  for(;ibtb != ibte;ibtb++){
	    if ( label[(*ibtb)->opposite->face] != i ){
	      neighborhoodBorder[i].push_back(*ibtb);
	    }
	  }
	  break;
	}
      }
      neighborhoodBorder_tmp[i].pop_front(); 
    }


    // we add the mesh border edge at the end
    std::list<HalfEdge*>::iterator 
      ictb = meshBorderEdge.begin(),
      icte = meshBorderEdge.end();
    for(; ictb != icte; ictb++){
      neighborhoodBorder[i].push_back((*ictb));
    }

    emptyNeighborhoods += (neighborhoodBorder[i].size() == 0);
	
    //// if neighborhood points empty, had the center (ex spikes)
    //if (neighborhoodPoints[i].size() == 0){
    //  neighborhoodPoints[i].push_back(i);
    //}

    // we reset the level two zero 
    // for the next vertex
    std::vector<int>::const_iterator 
      iptb = neighborhoodPoints[i].begin(), 
      ipte = neighborhoodPoints[i].end();
    for(; iptb != ipte; iptb++){
      level[(*iptb)] = 0;
    }   
  }
  
  if (emptyNeighborhoods > selection.size()/2.0){
    PRINT_ERROR("More than half the neighborhoods are empty, check your neighborhood size");
  }

  Stopwatch::Stop("Neighborhood"); 
}



std::vector<double> Neighborhood::ComputeSum(const std::vector<double>& field) const{

  int npoint = this->mesh->Points().size();  
  std::vector<double> values(npoint,0.0);
  
  for(int i=0; i<npoint; i++){
    std::vector<int>::const_iterator 
      itb = this->GetNeighborhoodPoints()[i].begin(), 
      ite = this->GetNeighborhoodPoints()[i].end();
    for(; itb != ite; itb++){
      values[i] += field[(*itb)];
    }
  }
  return values;
}

std::vector<double> Neighborhood::ComputeMean(const std::vector<double>& field) const{

  int npoint = this->mesh->Points().size();  
  std::vector<double> values(npoint,0.0);
  
  for(int i=0; i<npoint; i++){
    std::vector<int>::const_iterator 
      itb = this->GetNeighborhoodPoints()[i].begin(), 
      ite = this->GetNeighborhoodPoints()[i].end();
    for(; itb != ite; itb++){
      values[i] += field[(*itb)];      
    }
    if (this->GetNeighborhoodPoints()[i].size() !=0){
      values[i]/= this->GetNeighborhoodPoints()[i].size();
    }
  }
  return values;
}


void Neighborhood::Export(std::string str){
  std::ofstream out;
  out.open(str.c_str(), std::fstream::out);
  if(!out) { 
    PRINT_ERROR("Unable to open the file.");
  }
  out << "# idcenter PointNb p1 ... pn FaceNb f1 ... fn HedgeNb pa,pb pb,pc .... px,pz" << std::endl;
  
  for(unsigned int i=0; i<neighborhoodPoints.size(); i++){
    out << i << " ";
    out << neighborhoodPoints[i].size() << " ";
    for(unsigned int j=0; j<neighborhoodPoints[i].size(); j++){
      out << neighborhoodPoints[i][j] << " ";      
    }

    out << neighborhoodFaces[i].size()  << " ";
    for(unsigned int j=0; j<neighborhoodFaces[i].size(); j++){
      out << neighborhoodFaces[i][j] << " ";      
    }

    out << neighborhoodBorder[i].size() << " ";
    for(unsigned int j=0; j<neighborhoodBorder[i].size(); j++){
      out << neighborhoodBorder[i][j]->vertex << "," 
	   << neighborhoodBorder[i][j]->opposite->vertex << " ";      
    }
    out << std::endl;
  }
  out.close();
}

void Neighborhood::PrintStats(){
  std::vector<int> pnb;
  std::vector<int> bnb;
  std::vector<int> fnb;
  for(unsigned int i=0; i<neighborhoodPoints.size(); i++){
    pnb.push_back(neighborhoodPoints[i].size());
    bnb.push_back(neighborhoodBorder[i].size());
    fnb.push_back(neighborhoodFaces[i].size());
  }

  std::cout << "Neighborhood Statistics" << std::endl;
  double mean,dev;
  int min,max;
  GetMeanAndDev(pnb,mean,dev,min,max);
  std::cout << "Points" << std::endl;
  std::cout << " mean:" << mean<< std::endl;
  std::cout << " dev :" << dev<< std::endl;
  std::cout << " min :" << min<< std::endl;
  std::cout << " max :" << max<< std::endl;
  GetMeanAndDev(bnb,mean,dev,min,max);
  std::cout << "Borders" << std::endl;
  std::cout << " mean:" << mean<< std::endl;
  std::cout << " dev :" << dev<< std::endl;
  std::cout << " min :" << min<< std::endl;
  std::cout << " max :" << max<< std::endl;
  GetMeanAndDev(fnb,mean,dev,min,max);
  std::cout << "Faces" << std::endl;
  std::cout << " mean:" << mean<< std::endl;
  std::cout << " dev :" << dev<< std::endl;
  std::cout << " min :" << min<< std::endl;
  std::cout << " max :" << max<< std::endl;

}

