//
// Copyright (C) 2010  Clement Creusot
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "supportGenerator.h"

#include "byProduct.h"
#include "curvature/goldfeatherCubicSimple.h"
#include "descriptors/volume.h"
#include "descriptors/willmore.h"
#include "descriptors/distanceToLocalPlane.h"
#include "descriptors/LSD.h"
#include "geom.h"

#include "settings.h"

#include "common.h"
#include "stopwatch.h"


// Get the main support structure using the parameters defined in 'settings'




void SupportGenerator::
GetNeighborhoods(Mesh * mesh, 
		 std::map<std::string, Neighborhood*>& neighborhoods,
		 const std::vector<int>* subset){
  std::vector<std::map<std::string, std::string> >::const_iterator 
    itb = settings->GetNeighborhoodConfig().begin(), 
    ite = settings->GetNeighborhoodConfig().end();
  for( ; itb != ite; itb++){
    const std::map<std::string, std::string>& prop = *itb;
    std::string type = map_extract<std::string>(prop, "type");
    std::string size = map_extract<std::string>(prop, "size");
    std::string name = map_extract<std::string>(prop, "name");
    
    Neighborhood* neigh =NULL;
    if (type.compare("level") ==0){
      neigh = new Neighborhood(mesh, atoi(size.c_str()), subset);
    }
    else{
      neigh = new Neighborhood(mesh, atof(size.c_str()), subset);
    }
    neighborhoods[name] = neigh;
  }
}


void SupportGenerator::
GetFields(Mesh * mesh,
	  std::map<std::string, Neighborhood*>& neighborhoods,
	  std::map<std::string, std::vector<double> * >& scalars,
	  std::map<std::string, std::vector<Histogram*>* >& vectors,
	  const std::vector<int> * subset){
  std::vector<std::map<std::string, std::string> >::const_iterator 
    itb = settings->GetFieldConfig().begin(),
    ite = settings->GetFieldConfig().end();
  for( ; itb != ite; itb++){
    const std::map<std::string, std::string>& prop = *itb;
    std::string name = map_extract<std::string>(prop, "name");
    std::string type = map_extract<std::string>(prop, "type");
    std::string method = map_extract<std::string>(prop, "method");
    
    Neighborhood * neigh =NULL;
    if (map_has_key<std::string,std::string>(prop, "neighborhood")){
      std::string neighborhoodName = map_extract<std::string>(prop, 
							      "neighborhood");
      neigh = map_extract<Neighborhood*>(neighborhoods, neighborhoodName);
    }
    
    if (type.compare("curvature") ==0){
      
      std::string curvatureType("LIC");
      if (map_has_key<std::string,std::string>(prop, "curvatureType")){      
	curvatureType = map_extract<std::string>(prop,"curvatureType");
      }
      Curvature * curva;
      if (curvatureType.compare("yoshizawa") == 0 ){
	PRINT_ERROR("Yoshizawa method withdrawn (not open source).");
      }
      else{
	curva = new GoldfeatherCubicSimple(neigh) ;
	curva->SetMethodString(curvatureType);
      }
      curva->Initialise();
      curva->Compute(subset); 

  
      // FIXME avoid copy pass structure to constructeur
      scalars[append(name,"-k1")] = new std::vector<double>(curva->Getk1());
      scalars[append(name,"-k2")] = new std::vector<double>(curva->Getk2());
      scalars[append(name,"-b0")] = new std::vector<double>(curva->Getb0());
      scalars[append(name,"-b3")] = new std::vector<double>(curva->Getb3());
      
      const std::vector<double*>& cd1 = curva->Getd1();
      const std::vector<double*>& cd2 = curva->Getd2();
      const std::vector<double*>& cno = curva->GetNormals();
      std::vector<Histogram*>* d1 = 
	new std::vector<Histogram*>(cd1.size());
      std::vector<Histogram*>* d2 = 
	new std::vector<Histogram*>(cd2.size());
      std::vector<Histogram*>* normals = 
	new std::vector<Histogram*>(cno.size());
      for (unsigned int i=0; i<(*d1).size(); i++){
	(*d1)[i] = new HVector(cd1[i],3);	  
	(*d2)[i] = new HVector(cd2[i],3);
	(*normals)[i] = new HVector(cno[i],3);
      }

      vectors[append(name,"-d1")] = d1;
      vectors[append(name,"-d2")] = d2;
      vectors[append(name,"-normals")] = normals;
      
      delete curva;
    }
    else if (type.compare("normals") ==0){
      if (method.compare("localFaces") == 0){
        std::vector<double*> normals(mesh->Points().size());
        mesh->MakeNormals(normals);
        std::vector<Histogram*>* n = 
	  new std::vector<Histogram*>(mesh->Points().size());
        for (unsigned int i=0; i<n->size(); i++){
	  (*n)[i] = new HVector(normals[i],3);
        }      
        vectors[name] = n;      
      }
    }
    else if (type.compare("localSum") ==0){
      if (neigh==NULL){
	PRINT_ERROR("No given neighborhood for localSum");
      }
      if (subset != NULL){
	PRINT_ERROR("LocalSum not compatible with subset field computation");
      }
      std::string fieldName = map_extract<std::string>(prop, "field");
      std::vector<double>& field = (*map_extract<std::vector<double>*>(scalars, fieldName));
      scalars[name] = new std::vector<double>(neigh->ComputeSum(field));
    }
    else if (type.compare("localMean") ==0){
      if (neigh==NULL){
	PRINT_ERROR("No given neighborhood for localMean");
      }
      if (subset != NULL){
	PRINT_ERROR("LocalMean not compatible with subset field computation");
      }
      std::string fieldName = map_extract<std::string>(prop, "field");
      std::vector<double>& field = (*map_extract<std::vector<double>*>(scalars, fieldName));
      scalars[name] = new std::vector<double>(neigh->ComputeMean(field));
    }
    else if (type.compare("volume") ==0){
      Volume vol(neigh,subset);
      scalars[name] = new std::vector<double>(vol.GetVolume());
    }
    else if (type.compare("willmore") ==0){
      Willmore  willmore(mesh,subset);
      scalars[name] = new std::vector<double>(neigh->ComputeSum(willmore.GetValues()));
    }
    else if (type.compare("distToLocalPlane") ==0){
      std::string normalName = map_extract<std::string>(prop, "normal");
      std::vector<Histogram*>* normals = 
	map_extract<std::vector<Histogram*>* >(vectors,normalName);
      DistanceToLocalPlane DLP(neigh,normals,subset);
      scalars[name] = new std::vector<double>(DLP.GetDistanceToLocalPlane());
    }
    else if (type.compare("byProduct") ==0){
      std::vector<std::vector<double>*> items;
      std::string key("items");
      extract_list<std::vector<double>*>(prop,key, scalars, items);
      std::vector<double>* byproduct = new std::vector<double>;
      ByProduct BP(*byproduct);
      BP.Compute(method, items, prop);
      scalars[name] = byproduct;
    }
    else if (type.compare("symetry") ==0){
      if (subset != NULL){
	PRINT_ERROR("Symetry descriptor does not support subset computation yet");
      }
      std::string normalName = map_extract<std::string>(prop, "normal");
      std::vector<Histogram*>* normals = 
	map_extract<std::vector<Histogram*>* >(vectors,normalName);

      std::vector<Histogram*>* directions = NULL; // directions==NULL -> direction = (1,0,0)
      if (map_has_key<std::string,std::string>(prop,"startingDirection")){      
	std::string directionName = map_extract<std::string>(prop, "startingDirection");
	directions = map_extract<std::vector<Histogram*>* >(vectors,directionName);
      }

      std::vector<double>* values = NULL; // values == NULL -> use projection along normal (Z')
      if (map_has_key<std::string,std::string>(prop,"field")){      
	std::string valueName = map_extract<std::string>(prop, "field");
	values = map_extract<std::vector<double>* >(scalars,valueName);
      }
      double radiusBinSize = MapGet<double>(prop, "radiusBinSize");
      double angleBinNb = MapGet<double>(prop, "angleBinNb");
      LocalSymetryDescriptor LSD(neigh,radiusBinSize,angleBinNb,normals,directions,values);
      scalars[name] = new std::vector<double>(LSD.GetLSD());
      
    }
    else if (type.compare("matching") ==0){
      PRINT_ERROR("Descriptor matching code not present.");
    }
    else if (type.compare("read") ==0){
      if (subset != NULL){
	PRINT_ERROR("read support descriptor does not support subset computation yet");
      }
      if (method.compare("fromFile") == 0){
	std::string suffix = map_extract<std::string>(prop, "suffix");
	std::string filename = FindCorrespondingFile(settings->GetCurrentFilename(),suffix.c_str());
	std::vector<double>* field = new std::vector<double>;
	std::ifstream in;
	double value;
	std::cerr << filename << std::endl;
	in.open(filename.c_str());
	if(!in) { 
	  PRINT_ERROR("Unable to open the file %s",filename.c_str());
	}
	in >> value;
	while( !in.eof() && !in.fail()){ 
	  (*field).push_back(value);
	  in >> value;
	}
	if ((*field).size() != mesh->Points().size()){
	  PRINT_ERROR("Map have %d element, %d recquired", (int)(*field).size(),(int)mesh->Points().size());
	}
	scalars[name] = field;
      }
      else{
	PRINT_ERROR("Unrecognised method %s",method.c_str());
      }
    }
    else if (type.compare("adaboost") ==0){
      PRINT_ERROR("Adaboost matching not present");
    } 
    else if (type.compare("histogramProjection") ==0){
	PRINT_ERROR("Histogram projection code not present");
    }
    else if (type.compare("histogram") ==0){
      std::vector<int> selection ;
      if (subset==NULL){
	for (unsigned int id=0; id<mesh->Points().size(); id++){
	  selection.push_back(id);
	}
      }
      else{
	selection = (*subset);
      }  	
      
      Stopwatch::Start("histogram");
      std::vector<Histogram*>* histograms = 
	new std::vector<Histogram*>(mesh->Points().size());
      if (method.compare("spinImage") ==0){
	int radiusBinNb       = MapGet<int>(prop,"radiusBinNb");
	int heightBinNbOver   = MapGet<int>(prop,"heightBinNbOver");
	int heightBinNbUnder  = MapGet<int>(prop,"heightBinNbUnder");
	double radiusStepSize = MapGet<double>(prop,"radiusStepSize");
	double heightStepSize = MapGet<double>(prop,"heightStepSize");
	std::string normalName = map_extract<std::string>(prop,"normals");
	std::vector<Histogram*>& normals = 
	  *map_extract<std::vector<Histogram*>* >(vectors,normalName);

	std::vector<int>::iterator
	  itb = selection.begin(),
	  ite = selection.end();
	for(; itb!= ite; itb++){
	  unsigned int id= (*itb);
	  HMatrix * histogram = new HMatrix();
	  double * normal = dynamic_cast<HVector&>(*(normals[id])).GetRawData();
	  mesh->ComputeSpinImage(id, normal,
				 radiusBinNb,heightBinNbOver, heightBinNbUnder, 
				 radiusStepSize,heightStepSize, *histogram);
	  if (map_has_key<std::string,std::string>(prop,"normalizedTimes")){
	    double factor = MapGet<double>(prop,"normalizedTimes");
	    histogram->Normalize(factor);
	  }
	  (*histograms)[id] = histogram;
	}
      }
      else if (method.compare("ballonImage") ==0){
	int radiusBinNb       = MapGet<int>(prop,"radiusBinNb");
	double radiusStepSize = MapGet<double>(prop,"radiusStepSize");

	std::vector<int>::iterator
	  itb = selection.begin(),
	  ite = selection.end();
	for(; itb!= ite; itb++){
	  unsigned int id= (*itb);
	  HMatrix * histogram = new HMatrix();
	  mesh->ComputeBalloonImage(id,
				    radiusBinNb, 
				    radiusStepSize,
				    *histogram);
	  if (map_has_key<std::string,std::string>(prop,"normalizedTimes")){
	    double factor = MapGet<double>(prop,"normalizedTimes");
	    histogram->Normalize(factor);
	  }
	  (*histograms)[id] = histogram;
	}
      }
      else{
	PRINT_ERROR("Unknown method %s.", method.c_str());
      }
      vectors[name] = histograms;   
      Stopwatch::Stop("histogram");   
    }
    else{
      PRINT_ERROR("Unknown type %s.", type.c_str());
    }
  }
}

