//
// Copyright (C) 2010  Clement Creusot
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <fstream>
#include <limits>

// Provide a vtk viewer if VTK is installed.
#ifdef USE_VTK  
#include "vtkActor.h"
#include "vtkDoubleArray.h"
#include "vtkPointData.h"
#include "vtkProperty.h"
#include "vtkPolyData.h"
#include "vtkPolyDataMapper.h"
#include "vtkRenderer.h"
#include "vtkRenderWindow.h"
#include "vtkRenderWindowInteractor.h"
#include "vtkInteractorStyleTrackballCamera.h"
#include "vtkOBJReader.h"
#include "common/vtkCustom.h"
#endif 

#include "geometry/supportGenerator.h"

// Singleton class, defines global variable settings.
#include "common/settings.h"

// Initialize a global stopwatch for time measurements
#include "common/stopwatch.h"
STOPWATCH_INIT();

using namespace std;


/*******************************************************************************/

void Print_help( char* name){
  cerr << "Usage : " << name << " config.conf file.obj [fieldName]" << endl;
  cerr << "Options : " << endl;
  cerr << "  -i : interactive mode" << endl;
  exit(EXIT_FAILURE);
}


void
InterpretOptionsCustom(char* program,vector<pair<string,string> >& opt)
{
  vector<double> values;
  for (unsigned int i=0; i< opt.size(); i++){
    char c = opt[i].first[1]; 
    switch (c)
      {
      case 'i':
	cerr << "Interactive mode = ON" << endl;
	settings->interactiveMode = true;
	break;

      case 'h':
	Print_help(program);
	break;
	
      default:
	cerr <<  "Unknown option `-"<< c << "'." << endl;
	exit(EXIT_FAILURE);
      }
  }

}

/*******************************************************************************/


int main(int argc, char** argv)
{

  Stopwatch::ProgramStart();

  // Read the commandline and fill the Settings class
  vector<string> args;
  vector<pair<string,string> > opts;
  ParseCommandLine(argc, argv, "ih", args, opts);
  InterpretOptionsCustom(argv[0],opts);

  if(args.size() < 2){
    Print_help(argv[0]);
  }
  string configFile(args[0]);
  string filename(args[1]);
  settings->ReadConfig(configFile);

  if(args.size() != 3){
    cerr << "Fields available in the configuration files are " << StringJoin(ExtractKeys("name",settings->GetFields()), ",") << endl;
    Print_help(argv[0]);
  }  
  string fieldName(args[2]);



  vector<string> fieldNameTab;
  StringSplit<string>(fieldName,",",fieldNameTab);

  Stopwatch::Start("FieldGen");

  // Read the obj file
#ifdef USE_VTK
  //double radius=10.0;
  double scaleMin =-1;
  double scaleMax =-1;

  vtkOBJReader* dataOBJIn = vtkOBJReader::New();
  dataOBJIn->SetFileName(filename.c_str());
  dataOBJIn->Update();
  Mesh mesh(dataOBJIn->GetOutput());    
#else
  Mesh mesh(filename.c_str());    
#endif


  SupportGenerator supportGen;
  map<string, Neighborhood*> neighborhoods;
  map<string, vector<double> * > scalars;
  map<string, vector<Histogram*>* > vectors;

  // Computes the neighborhoods.
  supportGen.GetNeighborhoods(&mesh, neighborhoods);
				
  // Computes the maps.
  supportGen.GetFields(&mesh, neighborhoods, scalars, vectors);
    
  Stopwatch::Stop("FieldGen");


    
  // Write the scalar maps in files
  if( not settings->IsInteractive()){
    for (unsigned int i=0; i<fieldNameTab.size(); i++){
      string dname = fieldNameTab[i];
      string outputName = NoExtension(filename)+"."+dname+".map";
      if (map_has_key(scalars, dname)){
	cout << "Map file produced : " << outputName  <<endl;
	vector<double>& field = (*scalars[dname]);
	ofstream out;
	out.open(outputName.c_str(), fstream::out);
	if(!out) { 
	  PRINT_ERROR("Unable to open the file.");
	}

	cerr <<  settings->GetFieldConfig().size() << endl;
	out << field << endl;
	out.close();
      }
      else{
	PRINT_WARNING("%s is not a scalar field. No output produced.", dname.c_str());
      }
    }
  }
#ifdef USE_VTK
  // Show the result if the interactive mode is on
  else{
    if (fieldNameTab.size() != 0 ){	
      cerr << "Keyboard Commands:" << endl;
      cerr << "   q/e : quit "<< endl;
      cerr << "   w   : wire"<< endl;
      cerr << "   s   : surface"<< endl;
      cerr << "   f   : zoom"<< endl;
      cerr << "   u   : capture screen"<< endl;
      
      for (unsigned int i=0; i<fieldNameTab.size(); i++){
	string dname = fieldNameTab[i];
	cerr << dname << endl;
	vtkPolyDataMapper *objMapper = vtkPolyDataMapper::New();
	vtkActor *objActor = vtkActor::New();      
	vtkRenderer *ren1 = vtkRenderer::New();
	objMapper->SetInput(dataOBJIn->GetOutput());	
	objActor->SetMapper(objMapper);
	objActor->GetProperty()->SetOpacity(1.0);
	  
	ren1->SetBackground(1, 1, 1);
	ren1->AddActor(objActor);
	  
	if (! map_has_key(scalars,dname)){
	  //continue;
	  PRINT_ERROR("%s might not be a scalar field. It is not present in the list.",dname.c_str());
	}
	    
	vector<double>& field = (*scalars[dname]);
	//cerr << field.size() << endl;
	
	
	double mean, deviation, mini, maxi, Q1,med,Q3;
	double low,up;
	//GetMeanAndDev<double>(field, mean, deviation, mini, maxi); 
	GetStatFromList<double>(field,mean,deviation,mini,Q1,med,Q3,maxi);
	
	
	vtkDoubleArray * vk = vtkDoubleArray::New();
	for(int i=0; i<(int)field.size(); i++){
	  vk ->InsertTupleValue(i, &field[i]);    
	}
	vk->SetName("field");
	
	dataOBJIn->GetOutput()->GetPointData()->SetScalars(vk);
	dataOBJIn->GetOutput()->GetPointData()->SetActiveScalars("field"); 
	
	if(maxi - mini < 10*deviation){
	  low = mini; 
	  up = maxi; 
	}
	else{
	  low = mean - deviation/10; 
	  up = mean + deviation/10; 
	}
	low = max(mini,med - 3*(med-Q1));
	up =  maxi;//min(maxi,med + 3*(Q3-med));
	
	if (scaleMin != -1) low = scaleMin;
	if (scaleMax != -1) up = scaleMax;
	objMapper->ScalarVisibilityOn();    
	objMapper->SetScalarRange(low,up); 
      
      
	ren1->ResetCamera();
      
	vtkRenderWindow *renWin = vtkRenderWindow::New();
	renWin->AddRenderer(ren1);
	renWin->SetSize(500, 500);
      
	vtkRenderWindowInteractor *iren = vtkRenderWindowInteractor::New();
	iren->SetRenderWindow(renWin);

	vtkInteractorStyleTrackballCamera *style = 
	  vtkInteractorStyleTrackballCamera::New();  
	iren->SetInteractorStyle(style);

	// USER BUTTON CODE U
	vtkCallbackCommand* callbackButton = vtkCallbackCommand::New();
	callbackButton->SetCallback(handle_user_button_as_capture);
	iren->AddObserver(vtkCommand::UserEvent, callbackButton, 1.0);  
	// END USER BUTTON CODE

	renWin->Render(); 

	// This starts the event loop and as a side effect causes an initial render.
	iren->Start();
    

	objMapper->Delete();
	objActor->Delete();
	ren1->Delete();
	iren->Delete();
	style->Delete();
	renWin->Delete();
      } 
    }
  }
  dataOBJIn->Delete();
#else
  else{
    PRINT_ERROR("The VTK library is missing: no visualization.");
  }
#endif	        


  // Erase used memory
  map<string, Neighborhood*>::iterator 
    itnb = neighborhoods.begin(),
    itne = neighborhoods.end();
  for(;itnb!=itne; itnb++){
    delete (*itnb).second;
  }
  map<string, vector<double> * >::iterator
    itsb = scalars.begin(),
    itse = scalars.end();
  for(;itsb!=itse; itsb++){
    delete (*itsb).second;
  }
  map<string,vector<Histogram*> * >::iterator
    itvb = vectors.begin(),
    itve = vectors.end();
  for(;itvb!=itve; itvb++){
    for (unsigned int i = 0; i<(*itvb).second->size(); i++){
      delete (*(*itvb).second)[i];
    }
    delete (*itvb).second;
  }


  Stopwatch::ProgramStop();
  
  if( not settings->IsInteractive()){
    Stopwatch::PrintDetails();
  }

  return EXIT_SUCCESS;
}




